# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a polygon."""

from typing import TYPE_CHECKING

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity
from scipy.spatial.transform import Rotation as SpatialRotation

from ansys.geometry.core.math.matrix import Matrix33, Matrix44
from ansys.geometry.core.math.point import Point2D
from ansys.geometry.core.misc.checks import graphics_required
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv


class Polygon(SketchFace):
    """Provides for modeling regular polygons.

    Parameters
    ----------
    center: Point2D
        Center point of the circle.
    inner_radius : ~pint.Quantity | Distance | Real
        Inner radius (apothem) of the polygon.
    sides : int
        Number of sides of the polygon.
    angle : ~pint.Quantity | Angle | Real, default: 0
        Placement angle for orientation alignment.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        inner_radius: Quantity | Distance | Real,
        sides: int,
        angle: Quantity | Angle | Real = 0,
    ):
        """Initialize the polygon."""
        super().__init__()

        # Check the inputs
        self._center = center
        self._inner_radius = (
            inner_radius if isinstance(inner_radius, Distance) else Distance(inner_radius)
        )
        if self._inner_radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

        self._angle_offset = angle if isinstance(angle, Angle) else Angle(angle)

        # Verify that the number of sides is valid with preferred range
        if sides < 3:
            raise ValueError("The minimum number of sides to construct a polygon is three.")
        self._n_sides = sides

    @property
    def center(self) -> Point2D:
        """Center point of the polygon."""
        return self._center

    @property
    def inner_radius(self) -> Quantity:
        """Inner radius (apothem) of the polygon."""
        return self._inner_radius.value

    @property
    def n_sides(self) -> int:
        """Number of sides of the polygon."""
        return self._n_sides

    @property
    def angle(self) -> Quantity:
        """Orientation angle of the polygon."""
        return self._angle_offset.value

    @property
    def length(self) -> Quantity:
        """Side length of the polygon."""
        return 2 * self.inner_radius * np.tan(np.pi / self.n_sides)

    @property
    def outer_radius(self) -> Quantity:
        """Outer radius of the polygon."""
        return self.inner_radius / np.cos(np.pi / self.n_sides)

    @property
    def perimeter(self) -> Quantity:
        """Perimeter of the polygon."""
        return self.n_sides * self.length

    @property
    def area(self) -> Quantity:
        """Area of the polygon."""
        return (self.inner_radius * self.perimeter) / 2

    @property
    @graphics_required
    def visualization_polydata(self) -> "pv.PolyData":
        """VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        import pyvista as pv

        # Compensate z orientation by -np.pi / 2 to match Geometry service polygon processing
        rotation = Matrix33(
            SpatialRotation.from_euler(
                "xyz",
                [0, 0, -np.pi / 2 + self.angle.m_as(UNITS.radian)],
                degrees=False,
            ).as_matrix()
        )

        transformation_matrix = Matrix44(
            [
                [
                    rotation[0, 0],
                    rotation[0, 1],
                    rotation[0, 2],
                    self.center.x.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[1, 0],
                    rotation[1, 1],
                    rotation[1, 2],
                    self.center.y.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[2, 0],
                    rotation[2, 1],
                    rotation[2, 2],
                    0,
                ],
                [0, 0, 0, 1],
            ]
        )

        return pv.Polygon(
            radius=self.inner_radius.m_as(DEFAULT_UNITS.LENGTH),
            n_sides=self.n_sides,
        ).transform(transformation_matrix, inplace=True)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""PyAnsys Geometry is a Python wrapper for the Ansys Geometry service."""

# Version
# ------------------------------------------------------------------------------

import importlib.metadata as importlib_metadata

__version__ = importlib_metadata.version(__name__.replace(".", "-"))
"""PyAnsys Geometry version."""

# Ease import statements
# ------------------------------------------------------------------------------

import os

from ansys.geometry.core.connection.launcher import (
    launch_docker_modeler,
    launch_modeler,
    launch_modeler_with_core_service,
    launch_modeler_with_discovery,
    launch_modeler_with_discovery_and_pimlight,
    launch_modeler_with_geometry_service,
    launch_modeler_with_geometry_service_and_pimlight,
    launch_modeler_with_spaceclaim,
    launch_modeler_with_spaceclaim_and_pimlight,
    launch_remote_modeler,
)
from ansys.geometry.core.logger import LOG
from ansys.geometry.core.modeler import Modeler

# Global config constants
# ------------------------------------------------------------------------------
USE_SERVICE_COLORS: bool = False
"""Global constant for checking whether to use service colors for plotting
purposes. If set to False, the default colors will be used (speed gain)."""

# TODO: This is a deprecated constant. Not used in the codebase.
# Left here for backwards compatibility (with older versions of the codebase).
# https://github.com/ansys/pyansys-geometry/issues/1319
DISABLE_MULTIPLE_DESIGN_CHECK: bool = True
"""Deprecated constant. Use ``DISABLE_ACTIVE_DESIGN_CHECK`` instead."""

DISABLE_ACTIVE_DESIGN_CHECK: bool = False
"""Global constant for disabling the ``ensure_design_is_active`` check.

Only set this to false if you are sure you want to disable this check
and your objects will always exist on the server side.
"""

DOCUMENTATION_BUILD: bool = os.environ.get("PYANSYS_GEOMETRY_DOC_BUILD", "false").lower() == "true"
"""Global flag for the documentation to use the proper PyVista Jupyter backend."""



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for handling units homogeneously throughout PyAnsys Geometry."""

from beartype import beartype as check_input_types
from pint import Quantity, Unit, UnitRegistry, set_application_registry

from ansys.geometry.core.misc.checks import check_pint_unit_compatibility
from ansys.geometry.core.typing import Real

UNITS = UnitRegistry()
"""Units manager."""

# This forces pint to set the previous UnitRegistry as the one to use
set_application_registry(UNITS)


class PhysicalQuantity:
    """Provides the base class for handling units throughout PyAnsys Geometry.

    Parameters
    ----------
    unit : ~pint.Unit
        Units for the class.
    expected_dimensions : ~pint.Unit, default: None
        Units for the dimensionality of the physical quantity.
    """

    @check_input_types
    def __init__(self, unit: Unit, expected_dimensions: Unit | None = None):
        """Initialize the ``PhysicalQuantity`` class."""
        if expected_dimensions:
            check_pint_unit_compatibility(unit, expected_dimensions)

        self._unit = unit
        _, self._base_unit = UNITS.get_base_units(unit)

    @property
    def unit(self) -> Unit:
        """Unit of the object."""
        return self._unit

    @unit.setter
    @check_input_types
    def unit(self, unit: Unit) -> None:
        """Set the unit of the object."""
        check_pint_unit_compatibility(unit, self._base_unit)
        self._unit = unit

    @property
    def base_unit(self) -> Unit:
        """Base unit of the object."""
        return self._base_unit

    def _get_quantity(self, input: Real) -> Quantity:
        """Return the input value as a ~:class:`pint.Quantity` class.

        Parameters
        ----------
        input : Real
            Number to express as a quantity.

        Returns
        -------
        ~pint.Quantity
            Physical quantity that the number represents.
        """
        return Quantity(input, units=self.base_unit).to(self.unit)

    @check_input_types
    def _base_units_magnitude(self, input: Quantity) -> Real:
        """Get the input's :class:`pint.Quantity` magnitude in base units.

        Parameters
        ----------
        input : ~pint.Quantity
            :class:`pint.Quantity` class to process.

        Returns
        -------
        Real
            Input's magnitude in base units.
        """
        check_pint_unit_compatibility(input.units, self._base_unit)
        return input.to_base_units().m



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides geometry primitive representation for 2D and 3D points."""

from typing import TYPE_CHECKING, Union

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity, Unit

from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.misc.checks import check_ndarray_is_float_int, check_type
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.misc.units import UNITS, PhysicalQuantity
from ansys.geometry.core.typing import RealSequence

DEFAULT_POINT2D_VALUES = [np.nan, np.nan]
"""Default values for a 2D point."""

DEFAULT_POINT3D_VALUES = [np.nan, np.nan, np.nan]
"""Default values for a 3D point."""

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.math.vector import Vector2D, Vector3D

BASE_UNIT_LENGTH = UNITS.get_base_units(DEFAULT_UNITS.LENGTH)[1]
"""Default value for the length of the base unit."""


class Point2D(np.ndarray, PhysicalQuantity):
    """Provides geometry primitive representation for a 2D point.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence, default: DEFAULT_POINT2D_VALUES
        Direction arguments, either as a :class:`numpy.ndarray <numpy.ndarray>` class
        or as a ``RealSequence``.
    unit : ~pint.Unit | None, default: DEFAULT_UNITS.LENGTH
        Units for defining 2D point values. If not specified, the default unit is
        ``DEFAULT_UNITS.LENGTH``.
    """

    def __new__(
        cls,
        input: np.ndarray | RealSequence = DEFAULT_POINT2D_VALUES,
        unit: Unit | None = None,
    ):
        """Initialize the ``Point2D`` class."""
        # Build an empty np.ndarray object
        return np.zeros(len(input)).view(cls)

    def __init__(
        self,
        input: np.ndarray | RealSequence = DEFAULT_POINT2D_VALUES,
        unit: Unit | None = None,
    ):
        """Initialize the ``Point2D`` class."""
        # Call the PhysicalQuantity ctor
        unit = unit if unit else DEFAULT_UNITS.LENGTH
        super().__init__(unit, expected_dimensions=DEFAULT_UNITS.LENGTH)

        # Check the inputs
        (
            check_ndarray_is_float_int(input, "input")
            if isinstance(input, np.ndarray)
            else check_ndarray_is_float_int(np.asarray(input), "input")
        )

        # Check dimensions
        if len(input) != 2:
            raise ValueError("Point2D class must receive two arguments.")  # noqa: E501

        # Store values
        self._quantities = [Quantity(elem, units=unit) for elem in input]
        self.flat = [elem.to_base_units().m for elem in self._quantities]

    @check_input_types
    def __eq__(self, other: "Point2D") -> bool:
        """Equals operator for the ``Point2D`` class."""
        return np.array_equal(self, other)

    def __ne__(self, other: "Point2D") -> bool:
        """Not equals operator for the ``Point2D`` class."""
        return not self == other

    @check_input_types
    def __set_value(self, input: Quantity, idx: int) -> None:
        """General setter method for the ``Point2D`` class."""
        self[idx] = self._base_units_magnitude(input)
        self._quantities[idx] = input.to(self.unit)

    def __add__(self, other: Union["Point2D", "Vector2D"]) -> "Point2D":
        """Add operation for the ``Point2D`` class."""
        from ansys.geometry.core.math.vector import Vector2D

        check_type(other, (Point2D, Vector2D))
        point = np.add(self, other).view(Point2D)
        point._unit = self.unit
        point._base_unit = self.base_unit
        point._quantities = [np.nan, np.nan]
        return point

    def __sub__(self, other: "Point2D") -> "Point2D":
        """Subtraction operation for the ``Point2D`` class."""
        from ansys.geometry.core.math.vector import Vector2D

        check_type(other, (Point2D, Vector2D))
        point = np.subtract(self, other).view(Point2D)
        point._unit = self.unit
        point._base_unit = self.base_unit
        point._quantities = [np.nan, np.nan]
        return point

    @property
    def x(self) -> Quantity:
        """X plane component value."""
        if self._quantities[0] is np.nan:
            self._quantities[0] = Quantity(self[0], units=self.base_unit).to(self.unit)
        return self._quantities[0]

    @x.setter
    def x(self, x: Quantity) -> None:
        """Set the X plane component value."""
        self.__set_value(x, 0)

    @property
    def y(self) -> Quantity:
        """Y plane component value."""
        if self._quantities[1] is np.nan:
            self._quantities[1] = Quantity(self[1], units=self.base_unit).to(self.unit)
        return self._quantities[1]

    @y.setter
    def y(self, y: Quantity) -> None:
        """Set the Y plane component value."""
        self.__set_value(y, 1)

    @PhysicalQuantity.unit.getter
    def unit(self) -> Unit:
        """Get the unit of the object."""
        if hasattr(self, "_unit"):
            return self._unit
        else:
            self._unit = DEFAULT_UNITS.LENGTH
            return self._unit

    @PhysicalQuantity.base_unit.getter
    def base_unit(self) -> Unit:
        """Get the base unit of the object."""
        if hasattr(self, "_base_unit"):
            return self._base_unit
        else:
            self._base_unit = BASE_UNIT_LENGTH
            return self._base_unit


class Point3D(np.ndarray, PhysicalQuantity):
    """Provides geometry primitive representation for a 3D point.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence, default: DEFAULT_POINT3D_VALUES
        Direction arguments, either as a :class:`numpy.ndarray <numpy.ndarray>` class
        or as a ``RealSequence``.
    unit : ~pint.Unit | None, default: DEFAULT_UNITS.LENGTH
        Units for defining 3D point values. If not specified, the default unit is
        ``DEFAULT_UNITS.LENGTH``.
    """

    def __new__(
        cls,
        input: np.ndarray | RealSequence = DEFAULT_POINT3D_VALUES,
        unit: Unit | None = None,
    ):
        """Initialize ``Point3D`` class."""
        # Build an empty np.ndarray object
        return np.zeros(len(input)).view(cls)

    def __init__(
        self,
        input: np.ndarray | RealSequence = DEFAULT_POINT3D_VALUES,
        unit: Unit | None = None,
    ):
        """Initialize ``Point3D`` class."""
        # Call the PhysicalQuantity ctor
        unit = unit if unit else DEFAULT_UNITS.LENGTH
        super().__init__(unit, expected_dimensions=DEFAULT_UNITS.LENGTH)

        # Check the inputs
        (
            check_ndarray_is_float_int(input, "input")
            if isinstance(input, np.ndarray)
            else check_ndarray_is_float_int(np.asarray(input), "input")
        )

        # Check dimensions
        if len(input) != 3:
            raise ValueError("Point3D class must receive 3 arguments.")  # noqa: E501

        # Store values
        self._quantities = [Quantity(elem, units=unit) for elem in input]
        self.flat = [elem.to_base_units().m for elem in self._quantities]

    @check_input_types
    def __eq__(self, other: "Point3D") -> bool:
        """Equals operator for the ``Point3D`` class."""
        return np.array_equal(self, other)

    def __ne__(self, other: "Point3D") -> bool:
        """Not equals operator for the ``Point3D`` class."""
        return not self == other

    def __add__(self, other: Union["Point3D", "Vector3D"]) -> "Point3D":
        """Add operation for the ``Point3D`` class."""
        from ansys.geometry.core.math.vector import Vector3D

        check_type(other, (Point3D, Vector3D))
        point = np.add(self, other).view(Point3D)
        point._unit = self.unit
        point._base_unit = self.base_unit
        point._quantities = [np.nan, np.nan, np.nan]
        return point

    def __sub__(self, other: Union["Point3D", "Vector3D"]) -> "Point3D":
        """Subtraction operation for the ``Point3D`` class."""
        from ansys.geometry.core.math.vector import Vector3D

        check_type(other, (Point3D, Vector3D))
        point = np.subtract(self, other).view(Point3D)
        point._unit = self.unit
        point._base_unit = self.base_unit
        point._quantities = [np.nan, np.nan, np.nan]
        return point

    @check_input_types
    def __set_value(self, input: Quantity, idx: int) -> None:
        """General setter method for the ``Point3D`` class."""
        self[idx] = self._base_units_magnitude(input)
        self._quantities[idx] = input.to(self.unit)

    @property
    def x(self) -> Quantity:
        """X plane component value."""
        if self._quantities[0] is np.nan:
            self._quantities[0] = Quantity(self[0], units=self.base_unit).to(self.unit)
        return self._quantities[0]

    @x.setter
    def x(self, x: Quantity) -> None:
        """Set the X plane component value."""
        self.__set_value(x, 0)

    @property
    def y(self) -> Quantity:
        """Y plane component value."""
        if self._quantities[1] is np.nan:
            self._quantities[1] = Quantity(self[1], units=self.base_unit).to(self.unit)
        return self._quantities[1]

    @y.setter
    def y(self, y: Quantity) -> None:
        """Set the Y plane component value."""
        self.__set_value(y, 1)

    @property
    def z(self) -> Quantity:
        """Z plane component value."""
        if self._quantities[2] is np.nan:
            self._quantities[2] = Quantity(self[2], units=self.base_unit).to(self.unit)
        return self._quantities[2]

    @z.setter
    def z(self, z: Quantity) -> None:
        """Set the Z plane component value."""
        self.__set_value(z, 2)

    @PhysicalQuantity.unit.getter
    def unit(self) -> Unit:
        """Get the unit of the object."""
        if hasattr(self, "_unit"):
            return self._unit
        else:
            self._unit = DEFAULT_UNITS.LENGTH
            return self._unit

    @PhysicalQuantity.base_unit.getter
    def base_unit(self) -> Unit:
        """Get the base unit of the object."""
        if hasattr(self, "_base_unit"):
            return self._base_unit
        else:
            self._base_unit = BASE_UNIT_LENGTH
            return self._base_unit

    def transform(self, matrix: "Matrix44") -> "Point3D":
        """Transform the 3D point with a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4x4 transformation matrix to apply to the point.

        Returns
        -------
        Point3D
            New 3D point that is the transformed copy of the original 3D point after applying
            the transformation matrix.

        Notes
        -----
        Transform the ``Point3D`` object by applying the specified 4x4
        transformation matrix and return a new ``Point3D`` object representing the
        transformed point.
        """
        point_4x1 = np.append(self, 1)
        result_4x1 = matrix * point_4x1
        result_point = Point3D(result_4x1[0:3])
        result_point.unit = self.unit
        return result_point



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing components."""

from enum import Enum, unique
from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional, Union
import uuid

from beartype import beartype as check_input_types
from pint import Quantity

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.commands_pb2 import (
    CreateBeamSegmentsRequest,
    CreateDesignPointsRequest,
)
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from ansys.api.geometry.v0.components_pb2 import (
    CreateRequest,
    SetPlacementRequest,
    SetSharedTopologyRequest,
)
from ansys.api.geometry.v0.components_pb2_grpc import ComponentsStub
from ansys.api.geometry.v0.models_pb2 import Direction, Line, SetObjectNameRequest
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import (
    grpc_curve_to_curve,
    grpc_frame_to_frame,
    grpc_material_to_material,
    grpc_matrix_to_matrix,
    grpc_point_to_point3d,
    point3d_to_grpc_point,
    unit_vector_to_grpc_direction,
)
from ansys.geometry.core.designer.beam import (
    Beam,
    BeamCrossSectionInfo,
    BeamProfile,
    BeamProperties,
    SectionAnchorType,
)
from ansys.geometry.core.designer.body import Body, CollisionType, MasterBody
from ansys.geometry.core.designer.coordinate_system import CoordinateSystem
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.designer.face import Face
from ansys.geometry.core.designer.part import MasterComponent, Part
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.math.constants import IDENTITY_MATRIX44
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.checks import (
    ensure_design_is_active,
    graphics_required,
    min_backend_version,
)
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.options import TessellationOptions
from ansys.geometry.core.shapes.curves.circle import Circle
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.shapes.parameterization import Interval, ParamUV
from ansys.geometry.core.shapes.surfaces import TrimmedSurface
from ansys.geometry.core.sketch.sketch import Sketch
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from pyvista import MultiBlock, PolyData


@unique
class SharedTopologyType(Enum):
    """Shared topologies available."""

    SHARETYPE_NONE = 0
    SHARETYPE_SHARE = 1
    SHARETYPE_MERGE = 2
    SHARETYPE_GROUPS = 3


@unique
class ExtrusionDirection(Enum):
    """Enum for extrusion direction definition."""

    POSITIVE = "+"
    NEGATIVE = "-"

    @classmethod
    def from_string(cls, string: str, use_default_if_error: bool = False) -> "ExtrusionDirection":
        """Convert a string to an ``ExtrusionDirection`` enum."""
        lcase_string = string.lower()
        if lcase_string in ("+", "p", "pos", "positive"):
            return cls.POSITIVE
        elif lcase_string in ("-", "n", "neg", "negative"):
            return cls.NEGATIVE
        elif use_default_if_error:
            from ansys.geometry.core.logger import LOG

            LOG.warning("Invalid extrusion direction. Using default value (+).")
            return cls.POSITIVE
        else:  # pragma: no cover
            raise ValueError(f"Invalid extrusion direction: {string}.")

    def get_multiplier(self) -> int:
        """Get the multiplier for the extrusion direction.

        Returns
        -------
        int
            1 if the direction is positive, -1 if negative.
        """
        return 1 if self is ExtrusionDirection.POSITIVE else -1


class Component:
    """Provides for creating and managing a component.

    This class synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    name : str
        User-defined label for the new component.
    parent_component : Component or None
        Parent component to place the new component under within the design assembly. The
        default is ``None`` only when dealing with a ``Design`` object.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    template : Component, default: None
        Template to create this component from. This creates an
        instance component that shares a master with the template component.
    instance_name: str, default: None
        User defined optional name for the component instance.
    preexisting_id : str, default: None
        ID of a component pre-existing on the server side to use to create the component
        on the client-side data model. If an ID is specified, a new component is not
        created on the server.
    master_component : MasterComponent, default: None
        Master component to use to create a nested component instance instead
        of creating a new conponent.
    read_existing_comp : bool, default: False
        Whether an existing component on the service should be read. This
        parameter is only valid when connecting to an existing service session.
        Otherwise, avoid using this optional parameter.
    """

    # Types of the class instance private attributes
    _components: list["Component"]
    _beams: list[Beam]
    _coordinate_systems: list[CoordinateSystem]
    _design_points: list[DesignPoint]

    @protect_grpc
    @check_input_types
    def __init__(
        self,
        name: str,
        parent_component: Union["Component", None],
        grpc_client: GrpcClient,
        template: Optional["Component"] = None,
        instance_name: Optional[str] = None,
        preexisting_id: str | None = None,
        master_component: MasterComponent | None = None,
        read_existing_comp: bool = False,
    ):
        """Initialize the ``Component`` class."""
        # Initialize the client and stubs needed
        self._grpc_client = grpc_client
        self._component_stub = ComponentsStub(self._grpc_client.channel)
        self._commands_stub = CommandsStub(self._grpc_client.channel)

        # Align instance name behavior with the server - empty string if None
        instance_name = instance_name if instance_name else ""

        if preexisting_id:
            self._name = name
            self._id = preexisting_id
            self._instance_name = instance_name
        else:
            if parent_component:
                template_id = template.id if template else ""
                new_component = self._component_stub.Create(
                    CreateRequest(
                        name=name,
                        parent=parent_component.id,
                        template=template_id,
                        instance_name=instance_name,
                    )
                )

                # Remove this method call once we know Service sends correct ObjectPath id
                self._id = new_component.component.id
                self._name = new_component.component.name
                self._instance_name = new_component.component.instance_name
            else:
                self._name = name
                self._id = None
                self._instance_name = instance_name

        # Initialize needed instance variables
        self._components = []
        self._beams = []
        self._coordinate_systems = []
        self._design_points = []
        self._parent_component = parent_component
        self._is_alive = True
        self._shared_topology = None
        self._master_component = master_component

        # Populate client data model
        if template:
            # If this is not a nested instance
            if not master_component:
                # Create new MasterComponent, but use template's Part
                master = MasterComponent(
                    uuid.uuid4(),
                    f"master_{name}",
                    template._master_component.part,
                    template._master_component.transform,
                )
                self._master_component = master

            # Recurse - Create more children components from template's remaining children
            self.__create_children(template)

        elif not read_existing_comp:
            # This is an independent Component - Create new Part and MasterComponent
            p = Part(uuid.uuid4(), f"p_{name}", [], [])
            master = MasterComponent(uuid.uuid4(), f"master_{name}", p)
            self._master_component = master

        self._master_component.occurrences.append(self)

    def _clear_cached_bodies(self) -> None:
        """Clear the cached bodies."""
        if "bodies" in self.__dict__:
            del self.__dict__["bodies"]

    @property
    def id(self) -> str:
        """ID of the component."""
        return self._id

    @property
    def _grpc_id(self) -> EntityIdentifier:
        """ID of the component in gRPC format."""
        return EntityIdentifier(id=self.id)

    @property
    def name(self) -> str:
        """Name of the component."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set the name of the component."""
        self.set_name(value)

    @protect_grpc
    @check_input_types
    @min_backend_version(25, 2, 0)
    def set_name(self, name: str) -> None:
        """Set the name of the component."""
        self._grpc_client.log.debug(f"Renaming component {self.id} from '{self.name}' to '{name}'.")
        self._component_stub.SetName(SetObjectNameRequest(id=self._grpc_id, name=name))
        self._name = name

    @property
    def instance_name(self) -> str:
        """Name of the component instance."""
        return self._instance_name

    @property
    def components(self) -> list["Component"]:
        """List of ``Component`` objects inside of the component."""
        return self._components

    @cached_property
    def bodies(self) -> list[Body]:
        """List of ``Body`` objects inside of the component."""
        bodies = []
        for body in self._master_component.part.bodies:
            id = f"{self.id}/{body.id}" if self.parent_component else body.id
            if body.is_alive:
                bodies.append(Body(id, body.name, self, body))
        return bodies

    @property
    def beams(self) -> list[Beam]:
        """List of ``Beam`` objects inside of the component."""
        return self._beams

    @property
    def design_points(self) -> list[DesignPoint]:
        """List of ``DesignPoint`` objects inside of the component."""
        return self._design_points

    @property
    def coordinate_systems(self) -> list[CoordinateSystem]:
        """List of ``CoordinateSystem`` objects inside of the component."""
        return self._coordinate_systems

    @property
    def parent_component(self) -> "Component":
        """Parent of the component."""
        return self._parent_component

    @property
    def is_alive(self) -> bool:
        """Whether the component is still alive on the server side."""
        return self._is_alive

    @property
    def shared_topology(self) -> SharedTopologyType | None:
        """Shared topology type of the component (if any).

        Notes
        -----
        If no shared topology has been set, ``None`` is returned.
        """
        return self._shared_topology

    def __create_children(self, template: "Component") -> None:
        """Create new component and child bodies from ``template``."""
        for template_comp in template.components:
            new_id = self.id + "/" + template_comp.id.split("/")[-1]
            new = Component(
                template_comp.name,
                self,
                self._grpc_client,
                template=template_comp,
                preexisting_id=new_id,
                master_component=template_comp._master_component,
            )
            self.components.append(new)

    def get_all_bodies(self) -> list[Body]:
        """Get all bodies in the component hierarchy.

        Returns
        -------
        list[Body]
            List of all bodies in the component hierarchy.
        """
        bodies = []
        for comp in self.components:
            bodies.extend(comp.get_all_bodies())
        bodies.extend(self.bodies)
        return bodies

    def get_world_transform(self) -> Matrix44:
        """Get the full transformation matrix of the component in world space.

        Returns
        -------
        Matrix44
            4x4 transformation matrix of the component in world space.
        """
        if self.parent_component is None:
            return IDENTITY_MATRIX44
        return self.parent_component.get_world_transform() * self._master_component.transform

    @protect_grpc
    @ensure_design_is_active
    def modify_placement(
        self,
        translation: Vector3D | None = None,
        rotation_origin: Point3D | None = None,
        rotation_direction: UnitVector3D | None = None,
        rotation_angle: Quantity | Angle | Real = 0,
    ):
        """Apply a translation and/or rotation to the placement matrix.

        Parameters
        ----------
        translation : Vector3D, default: None
            Vector that defines the desired translation to the component.
        rotation_origin : Point3D, default: None
            Origin that defines the axis to rotate the component about.
        rotation_direction : UnitVector3D, default: None
            Direction of the axis to rotate the component about.
        rotation_angle : ~pint.Quantity | Angle | Real, default: 0
            Angle to rotate the component around the axis.

        Notes
        -----
        To reset a component's placement to an identity matrix, see
        :func:`reset_placement()` or call :func:`modify_placement()` with no arguments.
        """
        t = (
            Direction(x=translation.x, y=translation.y, z=translation.z)
            if translation is not None
            else None
        )
        p = point3d_to_grpc_point(rotation_origin) if rotation_origin is not None else None
        d = (
            unit_vector_to_grpc_direction(rotation_direction)
            if rotation_direction is not None
            else None
        )
        angle = rotation_angle if isinstance(rotation_angle, Angle) else Angle(rotation_angle)

        response = self._component_stub.SetPlacement(
            SetPlacementRequest(
                id=self.id,
                translation=t,
                rotation_axis_origin=p,
                rotation_axis_direction=d,
                rotation_angle=angle.value.m,
            )
        )
        self._master_component.transform = grpc_matrix_to_matrix(response.matrix)

    def reset_placement(self):
        """Reset a component's placement matrix to an identity matrix.

        See :func:`modify_placement()`.
        """
        self.modify_placement()

    @check_input_types
    @ensure_design_is_active
    def add_component(
        self, name: str, template: Optional["Component"] = None, instance_name: str = None
    ) -> "Component":
        """Add a new component under this component within the design assembly.

        Parameters
        ----------
        name : str
            User-defined label for the new component.
        template : Component, default: None
            Template to create this component from. This creates an
            instance component that shares a master with the template component.

        Returns
        -------
        Component
            New component with no children in the design assembly.
        """
        new_comp = Component(
            name, self, self._grpc_client, template=template, instance_name=instance_name
        )
        master = new_comp._master_component
        master_id = new_comp.id.split("/")[-1]
        for comp in self._master_component.occurrences:
            if comp.id != self.id:
                comp.components.append(
                    Component(
                        name,
                        comp,
                        self._grpc_client,
                        template=template,
                        instance_name=instance_name,
                        preexisting_id=f"{comp.id}/{master_id}",
                        master_component=master,
                        read_existing_comp=True,
                    )
                )

        self.components.append(new_comp)
        return self._components[-1]

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def set_shared_topology(self, share_type: SharedTopologyType) -> None:
        """Set the shared topology to apply to the component.

        Parameters
        ----------
        share_type : SharedTopologyType
            Shared topology type to assign to the component.
        """
        # Set the SharedTopologyType on the server
        self._grpc_client.log.debug(
            f"Setting shared topology type {share_type.value} on {self.id}."
        )
        self._component_stub.SetSharedTopology(
            SetSharedTopologyRequest(id=self.id, share_type=share_type.value)
        )

        # Store the SharedTopologyType set on the client
        self._shared_topology = share_type

    def __build_body_from_response(self, response: dict) -> Body:
        """Build a body from a response dictionary coming out of the gRPC call.

        Parameters
        ----------
        response : dict
            Response dictionary from the gRPC call.

        Returns
        -------
        Body
            Body object.

        Notes
        -----
        This is a completely private method and is intended to be
        used only within the class. It handles the MasterBody and
        Body creation, and addition to the component.
        """
        tb = MasterBody(
            response["master_id"],
            response["name"],
            self._grpc_client,
            is_surface=response["is_surface"],
        )
        self._master_component.part.bodies.append(tb)
        self._clear_cached_bodies()
        return Body(response["id"], response["name"], self, tb)

    @check_input_types
    @ensure_design_is_active
    def extrude_sketch(
        self,
        name: str,
        sketch: Sketch,
        distance: Quantity | Distance | Real,
        direction: ExtrusionDirection | str = ExtrusionDirection.POSITIVE,
        cut: bool = False,
    ) -> Body | None:
        """Create a solid body by extruding the sketch profile a distance.

        Parameters
        ----------
        name : str
            User-defined label for the new solid body.
        sketch : Sketch
            Two-dimensional sketch source for the extrusion.
        distance : ~pint.Quantity | Distance | Real
            Distance to extrude the solid body.
        direction : ExtrusionDirection | str, default: "+"
            Direction for extruding the solid body.
            The default is to extrude in the positive normal direction of the sketch.
            Options are "+" and "-" as a string, or the enum values.
        cut : bool, default: False
            Whether to cut the extrusion from the existing component.
            By default, the extrusion is added to the existing component.

        Returns
        -------
        Body
            Extruded body from the given sketch.
        None
            If the cut parameter is ``True``, the function returns ``None``.

        Notes
        -----
        The newly created body is placed under this component within the design assembly.
        """
        # Sanity checks on inputs
        distance = distance if isinstance(distance, Distance) else Distance(distance)
        if isinstance(direction, str):
            direction = ExtrusionDirection.from_string(direction, use_default_if_error=True)

        # Perform extrusion request
        self._grpc_client.log.debug(f"Extruding sketch provided on {self.id}. Creating body...")
        response = self._grpc_client.services.bodies.create_extruded_body(
            name=name,
            parent_id=self.id,
            sketch=sketch,
            distance=distance,
            direction=direction.get_multiplier(),
        )
        created_body = self.__build_body_from_response(response)

        if not cut:
            return created_body
        else:
            # If cut is True, subtract the created body from all existing bodies
            # in the component...
            for existing_body in self.get_all_bodies():
                # Skip the created body
                if existing_body.id == created_body.id:
                    continue
                # Check for collision
                if existing_body.get_collision(created_body) != CollisionType.NONE:
                    existing_body.subtract(created_body, keep_other=True)

            # Finally, delete the created body
            self.delete_body(created_body)

            # And obviously return None... since no body is created
            return None

    @min_backend_version(24, 2, 0)
    @check_input_types
    @ensure_design_is_active
    def sweep_sketch(
        self,
        name: str,
        sketch: Sketch,
        path: list[TrimmedCurve],
    ) -> Body:
        """Create a body by sweeping a planar profile along a path.

        Parameters
        ----------
        name : str
            User-defined label for the new solid body.
        sketch : Sketch
            Two-dimensional sketch source for the extrusion.
        path : list[TrimmedCurve]
            The path to sweep the profile along.

        Returns
        -------
        Body
            Created body from the given sketch.

        Notes
        -----
        The newly created body is placed under this component within the design assembly.
        """
        self._grpc_client.log.debug(f"Creating a sweeping profile on {self.id}. Creating body...")
        response = self._grpc_client.services.bodies.create_sweeping_profile_body(
            name=name,
            parent_id=self.id,
            sketch=sketch,
            path=path,
        )
        return self.__build_body_from_response(response)

    @min_backend_version(24, 2, 0)
    @check_input_types
    @ensure_design_is_active
    def sweep_chain(
        self,
        name: str,
        path: list[TrimmedCurve],
        chain: list[TrimmedCurve],
    ) -> Body:
        """Create a body by sweeping a chain of curves along a path.

        Parameters
        ----------
        name : str
            User-defined label for the new solid body.
        path : list[TrimmedCurve]
            The path to sweep the chain along.
        chain : list[TrimmedCurve]
            A chain of trimmed curves.

        Returns
        -------
        Body
            Created body from the given sketch.

        Notes
        -----
        The newly created body is placed under this component within the design assembly.
        """
        self._grpc_client.log.debug(f"Creating a sweeping chain on {self.id}. Creating body...")
        response = self._grpc_client.services.bodies.create_sweeping_chain(
            name=name,
            parent_id=self.id,
            path=path,
            chain=chain,
        )
        return self.__build_body_from_response(response)

    @min_backend_version(24, 2, 0)
    @check_input_types
    def revolve_sketch(
        self,
        name: str,
        sketch: Sketch,
        axis: Vector3D,
        angle: Quantity | Angle | Real,
        rotation_origin: Point3D,
    ) -> Body:
        """Create a solid body by revolving a sketch profile around an axis.

        Parameters
        ----------
        name : str
            User-defined label for the new solid body.
        sketch : Sketch
            Two-dimensional sketch source for the revolve.
        axis : Vector3D
            Axis of rotation for the revolve.
        angle : ~pint.Quantity | Angle | Real
            Angle to revolve the solid body around the axis. The angle can be positive or negative.
        rotation_origin : Point3D
            Origin of the axis of rotation.

        Returns
        -------
        Body
            Revolved body from the given sketch.
        """
        # Based on the reference axis and the sketch plane's normal, retrieve the orthogonal
        # vector (i.e. this is the reference vector for the Circle object). Assuming a distance of 1
        # we revolve around the axis the angle given.
        rotation_vector = sketch._plane.normal.cross(axis)

        # Define the revolve path
        circle = Circle(
            rotation_origin,
            radius=Distance(1),
            reference=rotation_vector,
            axis=axis,
        )
        angle = angle if isinstance(angle, Angle) else Angle(angle)
        interval = (
            Interval(0, angle.value.m_as(DEFAULT_UNITS.SERVER_ANGLE))
            if angle.value.m >= 0
            else Interval(angle.value.m_as(DEFAULT_UNITS.SERVER_ANGLE), 0)
        )
        path = circle.trim(interval)

        # Create the revolved body by delegating to the sweep method
        return self.sweep_sketch(name, sketch, [path])

    @check_input_types
    @ensure_design_is_active
    def extrude_face(
        self,
        name: str,
        face: Face,
        distance: Quantity | Distance,
        direction: ExtrusionDirection | str = ExtrusionDirection.POSITIVE,
    ) -> Body:
        """Extrude the face profile by a given distance to create a solid body.

        There are no modifications against the body containing the source face.

        Parameters
        ----------
        name : str
            User-defined label for the new solid body.
        face : Face
            Target face to use as the source for the new surface.
        distance : ~pint.Quantity | Distance | Real
            Distance to extrude the solid body.
        direction : ExtrusionDirection | str, default: "+"
            Direction for extruding the solid body's face.
            The default is to extrude in the positive normal direction of the face.
            Options are "+" and "-" as a string, or the enum values.

        Returns
        -------
        Body
            Extruded solid body.

        Notes
        -----
        The source face can be anywhere within the design component hierarchy.
        Therefore, there is no validation requiring that the face is placed under the
        target component where the body is to be created.
        """
        # Sanity checks on inputs
        distance = distance if isinstance(distance, Distance) else Distance(distance)
        if isinstance(direction, str):
            direction = ExtrusionDirection.from_string(direction, use_default_if_error=True)

        self._grpc_client.log.debug(f"Extruding from face provided on {self.id}. Creating body...")
        response = self._grpc_client.services.bodies.create_extruded_body_from_face_profile(
            name=name,
            parent_id=self.id,
            face_id=face.id,
            distance=distance,
            direction=direction.get_multiplier(),
        )

        return self.__build_body_from_response(response)

    @check_input_types
    @ensure_design_is_active
    @min_backend_version(24, 2, 0)
    def create_sphere(self, name: str, center: Point3D, radius: Distance) -> Body:
        """Create a sphere body defined by the center point and the radius.

        Parameters
        ----------
        name : str
            Body name.
        center : Point3D
            Center point of the sphere.
        radius : Distance
            Radius of the sphere.

        Returns
        -------
        Body
            Sphere body object.
        """
        self._grpc_client.log.debug(f"Creating a sphere body on {self.id}.")
        response = self._grpc_client.services.bodies.create_sphere_body(
            name=name, parent=self.id, center=center, radius=radius
        )
        return self.__build_body_from_response(response)

    @check_input_types
    @ensure_design_is_active
    @min_backend_version(24, 2, 0)
    def create_body_from_loft_profile(
        self,
        name: str,
        profiles: list[list[TrimmedCurve]],
        periodic: bool = False,
        ruled: bool = False,
    ) -> Body:
        """Create a lofted body from a collection of trimmed curves.

        Parameters
        ----------
        name : str
            Name of the lofted body.
        profiles : list[list[TrimmedCurve]]
            Collection of lists of trimmed curves (profiles) defining the lofted body's shape.
        periodic : bool, default: False
            Whether the lofted body should have periodic continuity.
        ruled : bool
            Whether the lofted body should be ruled.

        Returns
        -------
        Body
            Created lofted body object.

        Notes
        -----
        Surfaces produced have a U parameter in the direction of the profile curves,
        and a V parameter in the direction of lofting.
        Profiles can have different numbers of segments. A minimum twist solution is
        produced.
        Profiles should be all closed or all open. Closed profiles cannot contain inner
        loops. If closed profiles are supplied, a closed (solid) body is produced, if
        possible. Otherwise, an open (sheet) body is produced.
        The periodic argument applies when the profiles are closed. It is ignored if
        the profiles are open.

        If ``periodic=True``, at least three profiles must be supplied. The loft continues
        from the last profile back to the first profile to produce surfaces that are
        periodic in V.

        If ``periodic=False``, at least two profiles must be supplied. If the first
        and last profiles are planar, end capping faces are created. Otherwise, an open
        (sheet) body is produced.
        If ``ruled=True``, separate ruled surfaces are produced between each pair of profiles.
        If ``periodic=True``, the loft continues from the last profile back to the first
        profile, but the surfaces are not periodic.
        """
        self._grpc_client.log.debug(f"Creating a loft profile body on {self.id}.")
        response = self._grpc_client.services.bodies.create_extruded_body_from_loft_profiles(
            name=name,
            parent_id=self.id,
            profiles=profiles,
            periodic=periodic,
            ruled=ruled,
        )
        return self.__build_body_from_response(response)

    @check_input_types
    @ensure_design_is_active
    def create_surface(self, name: str, sketch: Sketch) -> Body:
        """Create a surface body with a sketch profile.

        The newly created body is placed under this component within the design assembly.

        Parameters
        ----------
        name : str
            User-defined label for the new surface body.
        sketch : Sketch
            Two-dimensional sketch source for the surface definition.

        Returns
        -------
        Body
            Body (as a planar surface) from the given sketch.
        """
        self._grpc_client.log.debug(
            f"Creating planar surface from sketch provided on {self.id}. Creating body..."
        )
        response = self._grpc_client.services.bodies.create_planar_body(
            name=name, parent_id=self.id, sketch=sketch
        )

        return self.__build_body_from_response(response)

    @check_input_types
    @ensure_design_is_active
    def create_surface_from_face(self, name: str, face: Face) -> Body:
        """Create a surface body based on a face.

        Parameters
        ----------
        name : str
            User-defined label for the new surface body.
        face : Face
            Target face to use as the source for the new surface.

        Returns
        -------
        Body
            Surface body.

        Notes
        -----
        The source face can be anywhere within the design component hierarchy.
        Therefore, there is no validation requiring that the face is placed under the
        target component where the body is to be created.
        """
        self._grpc_client.log.debug(
            f"Creating planar surface from face provided on {self.id}. Creating body..."
        )
        response = self._grpc_client.services.bodies.create_body_from_face(
            name=name, parent_id=self.id, face_id=face.id
        )
        return self.__build_body_from_response(response)

    @check_input_types
    @ensure_design_is_active
    @min_backend_version(25, 1, 0)
    def create_body_from_surface(self, name: str, trimmed_surface: TrimmedSurface) -> Body:
        """Create a surface body from a trimmed surface.

        Parameters
        ----------
        name : str
            User-defined label for the new surface body.
        trimmed_surface : TrimmedSurface
            Geometry for the new surface body.

        Returns
        -------
        Body
            Surface body.

        Notes
        -----
        It is possible to create a closed solid body (as opposed to an open surface body) with a
        Sphere or Torus if they are untrimmed. This can be validated with `body.is_surface`.
        """
        self._grpc_client.log.debug(
            f"Creating surface body from trimmed surface provided on {self.id}. Creating body..."
        )
        response = self._grpc_client.services.bodies.create_surface_body(
            name=name,
            parent_id=self.id,
            trimmed_surface=trimmed_surface,
        )
        return self.__build_body_from_response(response)

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def create_surface_from_trimmed_curves(
        self, name: str, trimmed_curves: list[TrimmedCurve]
    ) -> Body:
        """Create a surface body from a list of trimmed curves all lying on the same plane.

        Parameters
        ----------
        name : str
            User-defined label for the new surface body.
        trimmed_curves : list[TrimmedCurve]
            Curves to define the plane and body.

        Returns
        -------
        Body
            Surface body.
        """
        self._grpc_client.log.debug(
            f"Creating surface body from trimmed curves provided on {self.id}. Creating body..."
        )
        response = self._grpc_client.services.bodies.create_surface_body_from_trimmed_curves(
            name=name,
            parent_id=self.id,
            trimmed_curves=trimmed_curves,
        )
        return self.__build_body_from_response(response)

    @check_input_types
    @ensure_design_is_active
    def create_coordinate_system(self, name: str, frame: Frame) -> CoordinateSystem:
        """Create a coordinate system.

        The newly created coordinate system is place under this component
        within the design assembly.

        Parameters
        ----------
        name : str
            User-defined label for the new coordinate system.
        frame : Frame
            Frame defining the coordinate system bounds.

        Returns
        -------
        CoordinateSystem
        """
        self._coordinate_systems.append(CoordinateSystem(name, frame, self, self._grpc_client))
        return self._coordinate_systems[-1]

    @check_input_types
    @ensure_design_is_active
    def translate_bodies(
        self, bodies: list[Body], direction: UnitVector3D, distance: Quantity | Distance | Real
    ) -> None:
        """Translate the bodies in a specified direction by a distance.

        Parameters
        ----------
        bodies: list[Body]
            list of bodies to translate by the same distance.
        direction: UnitVector3D
            Direction of the translation.
        distance: ~pint.Quantity | Distance | Real
            Magnitude of the translation.

        Returns
        -------
        None

        Notes
        -----
        If the body does not belong to this component (or its children), it
        is not translated.
        """
        body_ids_found = []

        for body in bodies:
            body_requested = self.search_body(body.id)
            if body_requested:
                body_ids_found.append(body_requested.id)
            else:
                self._grpc_client.log.warning(
                    f"Body with ID {body.id} and name {body.name} is not found in this "
                    + "component (or subcomponents). Ignoring this translation request."
                )
                pass

        distance = distance if isinstance(distance, Distance) else Distance(distance)

        self._grpc_client.log.debug(f"Translating {body_ids_found}...")
        self._grpc_client.services.bodies.translate(
            ids=body_ids_found,
            direction=direction,
            distance=distance,
        )

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def create_beams(
        self,
        segments: list[tuple[Point3D, Point3D]],
        profile: BeamProfile,
    ) -> list[Beam]:
        """Create beams under the component.

        Parameters
        ----------
        segments : list[tuple[Point3D, Point3D]]
            List of start and end pairs, each specifying a single line segment.
        profile : BeamProfile
            Beam profile to use to create the beams.

        Returns
        -------
        list[Beam]
            A list of the created Beams.

        Notes
        -----
        The newly created beams synchronize to a design within a supporting
        Geometry service instance.
        """
        if self._grpc_client.backend_version < (25, 2, 0):
            return self.__create_beams_legacy(segments, profile)
        else:
            return self.__create_beams(segments, profile)

    def __create_beams_legacy(
        self, segments: list[tuple[Point3D, Point3D]], profile: BeamProfile
    ) -> list[Beam]:
        """Create beams under the component.

        Parameters
        ----------
        segments : list[tuple[Point3D, Point3D]]
            List of start and end pairs, each specifying a single line segment.
        profile : BeamProfile
            Beam profile to use to create the beams.

        Returns
        -------
        list[Beam]
            A list of the created Beams.

        Notes
        -----
        This is a legacy method, which is used in versions up to Ansys 25.1.1 products.
        """
        request = CreateBeamSegmentsRequest(parent=self.id, profile=profile.id)

        for segment in segments:
            request.lines.append(
                Line(start=point3d_to_grpc_point(segment[0]), end=point3d_to_grpc_point(segment[1]))
            )

        self._grpc_client.log.debug(f"Creating beams on {self.id}...")
        response = self._commands_stub.CreateBeamSegments(request)
        self._grpc_client.log.debug("Beams successfully created.")

        # Note: The current gRPC API simply returns a list of IDs. There is no additional
        # information to correlate/merge against, so it is fully assumed that the list is
        # returned in order with a 1 to 1 index match to the request segments list.
        new_beams = []
        n_beams = len(response.ids)
        for index in range(n_beams):
            new_beams.append(
                Beam(response.ids[index], segments[index][0], segments[index][1], profile, self)
            )

        self._beams.extend(new_beams)
        return self._beams[-n_beams:]

    def __create_beams(
        self,
        segments: list[tuple[Point3D, Point3D]],
        profile: BeamProfile,
    ) -> list[Beam]:
        """Create beams under the component.

        Parameters
        ----------
        segments : list[tuple[Point3D, Point3D]]
            List of start and end pairs, each specifying a single line segment.
        profile : BeamProfile
            Beam profile to use to create the beams.

        Returns
        -------
        list[Beam]
            A list of the created Beams.
        """
        request = CreateBeamSegmentsRequest(
            profile=profile.id,
            parent=self.id,
        )

        for segment in segments:
            request.lines.append(
                Line(start=point3d_to_grpc_point(segment[0]), end=point3d_to_grpc_point(segment[1]))
            )

        self._grpc_client.log.debug(f"Creating beams on {self.id}...")
        response = self._commands_stub.CreateDescriptiveBeamSegments(request)
        self._grpc_client.log.debug("Beams successfully created.")

        beams = []
        for beam in response.created_beams:
            cross_section = BeamCrossSectionInfo(
                section_anchor=SectionAnchorType(beam.cross_section.section_anchor),
                section_angle=beam.cross_section.section_angle,
                section_frame=grpc_frame_to_frame(beam.cross_section.section_frame),
                section_profile=[
                    [
                        TrimmedCurve(
                            geometry=grpc_curve_to_curve(curve.geometry),
                            start=grpc_point_to_point3d(curve.start),
                            end=grpc_point_to_point3d(curve.end),
                            interval=Interval(curve.interval_start, curve.interval_end),
                            length=curve.length,
                        )
                        for curve in curve_list
                    ]
                    for curve_list in beam.cross_section.section_profile
                ],
            )
            properties = BeamProperties(
                area=beam.properties.area,
                centroid=ParamUV(beam.properties.centroid_x, beam.properties.centroid_y),
                warping_constant=beam.properties.warping_constant,
                ixx=beam.properties.ixx,
                ixy=beam.properties.ixy,
                iyy=beam.properties.iyy,
                shear_center=ParamUV(
                    beam.properties.shear_center_x, beam.properties.shear_center_y
                ),
                torsion_constant=beam.properties.torsional_constant,
            )

            beams.append(
                Beam(
                    id=beam.id.id,
                    start=grpc_point_to_point3d(beam.shape.start),
                    end=grpc_point_to_point3d(beam.shape.end),
                    profile=profile,
                    parent_component=self,
                    name=beam.name,
                    is_deleted=beam.is_deleted,
                    is_reversed=beam.is_reversed,
                    is_rigid=beam.is_rigid,
                    material=grpc_material_to_material(beam.material),
                    cross_section=cross_section,
                    properties=properties,
                    shape=beam.shape,
                    beam_type=beam.type,
                )
            )

        self._beams.extend(beams)
        return beams

    def create_beam(self, start: Point3D, end: Point3D, profile: BeamProfile) -> Beam:
        """Create a beam under the component.

        The newly created beam synchronizes to a design within a supporting
        Geometry service instance.

        Parameters
        ----------
        start : Point3D
            Starting point of the beam line segment.
        end : Point3D
            Ending point of the beam line segment.
        profile : BeamProfile
            Beam profile to use to create the beam.
        """
        return self.create_beams([(start, end)], profile)[0]

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def delete_component(self, component: Union["Component", str]) -> None:
        """Delete a component (itself or its children).

        Parameters
        ----------
        component : Component | str
            ID of the component or instance to delete.

        Notes
        -----
        If the component is not this component (or its children), it
        is not deleted.
        """
        id = component if isinstance(component, str) else component.id
        component_requested = self.search_component(id)

        if component_requested:
            # If the component belongs to this component (or nested components)
            # call the server deletion mechanism
            self._component_stub.Delete(EntityIdentifier(id=id))

            # If the component was deleted from the server side... "kill" it
            # on the client side
            component_requested._kill_component_on_client()
            self._grpc_client.log.debug(f"Component {component_requested.id} has been deleted.")
        else:
            self._grpc_client.log.warning(
                f"Component {id} not found in this component (or subcomponents)."
                + " Ignoring deletion request."
            )
            pass

    @check_input_types
    @ensure_design_is_active
    def delete_body(self, body: Body | str) -> None:
        """Delete a body belonging to this component (or its children).

        Parameters
        ----------
        body : Body | str
            ID of the body or instance to delete.

        Notes
        -----
        If the body does not belong to this component (or its children), it
        is not deleted.
        """
        id = body if isinstance(body, str) else body.id
        body_requested = self.search_body(id)

        if body_requested:
            # If the body belongs to this component (or nested components)
            # call the server deletion mechanism
            self._grpc_client.services.bodies.delete(id=id)

            # If the body was deleted from the server side... "kill" it
            # on the client side
            body_requested._is_alive = False
            self._grpc_client.log.debug(f"Body {body_requested.id} has been deleted.")
            self._clear_cached_bodies()
        else:
            self._grpc_client.log.warning(
                f"Body {id} is not found in this component (or subcomponents)."
                + " Ignoring this deletion request."
            )
            pass

    def add_design_point(
        self,
        name: str,
        point: Point3D,
    ) -> DesignPoint:
        """Create a single design point.

        Parameters
        ----------
        name : str
            User-defined label for the design points.
        points : Point3D
            3D point constituting the design point.
        """
        return self.add_design_points(name, [point])[0]

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def add_design_points(
        self,
        name: str,
        points: list[Point3D],
    ) -> list[DesignPoint]:
        """Create a list of design points.

        Parameters
        ----------
        name : str
            User-defined label for the list of design points.
        points : list[Point3D]
            list of the 3D points that constitute the list of design points.
        """
        # Create DesignPoint objects server-side
        self._grpc_client.log.debug(f"Creating design points on {self.id}...")
        response = self._commands_stub.CreateDesignPoints(
            CreateDesignPointsRequest(
                points=[point3d_to_grpc_point(point) for point in points], parent=self.id
            )
        )
        self._grpc_client.log.debug("Design points successfully created.")

        # Once created on the server, create them client side
        new_design_points = []
        n_design_points = len(response.ids)
        for index in range(n_design_points):
            new_design_points.append((DesignPoint(response.ids[index], name, points[index], self)))
        self._design_points.extend(new_design_points)

        # Finally return the list of created DesignPoint objects
        return self._design_points[-n_design_points:]

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def delete_beam(self, beam: Beam | str) -> None:
        """Delete an existing beam belonging to this component's scope.

        Parameters
        ----------
        beam : Beam | str
            ID of the beam or instance to delete.

        Notes
        -----
        If the beam belongs to this component's children, it is deleted.
        If the beam does not belong to this component (or its children), it
        is not deleted.
        """
        id = beam if isinstance(beam, str) else beam.id
        beam_requested = self.search_beam(id)

        if beam_requested:
            # If the beam belongs to this component (or nested components)
            # call the server deletion mechanism
            #
            # Server-side, the same deletion request has to be performed
            # as for deleting a Body
            #
            self._commands_stub.DeleteBeam(EntityIdentifier(id=beam_requested.id))

            # If the beam was deleted from the server side... "kill" it
            # on the client side
            beam_requested._is_alive = False
            self._grpc_client.log.debug(f"Beam {beam_requested.id} has been deleted.")
        else:
            self._grpc_client.log.warning(
                f"Beam {id} not found in this component (or subcomponents)."
                + " Ignoring deletion request."
            )
            pass

    @check_input_types
    def search_component(self, id: str) -> Union["Component", None]:
        """Search nested components recursively for a component.

        Parameters
        ----------
        id : str
            ID of the component to search for.

        Returns
        -------
        Component
           Component with the requested ID. If this ID is not found, ``None`` is returned.
        """
        # Check if the requested component is this one
        if self.id == id and self.is_alive:
            return self

        # If no luck, search on nested components
        result = None
        for component in self.components:
            result = component.search_component(id)
            if result:
                return result

        # If you reached this point... this means that no component was found!
        return None

    @check_input_types
    def search_body(self, id: str) -> Body | None:
        """Search bodies in the component's scope.

        Parameters
        ----------
        id : str
            ID of the body to search for.

        Returns
        -------
        Body | None
            Body with the requested ID. If the ID is not found, ``None`` is returned.

        Notes
        -----
        This method searches for bodies in the component and nested components
        recursively.
        """
        # Search in component's bodies
        for body in self.bodies:
            if body.id == id and body.is_alive:
                return body

        # If no luck, search on nested components
        result = None
        for component in self.components:
            result = component.search_body(id)
            if result:
                return result

        # If you reached this point... this means that no body was found!
        return None

    @check_input_types
    def search_beam(self, id: str) -> Beam | None:
        """Search beams in the component's scope.

        Parameters
        ----------
        id : str
            ID of the beam to search for.

        Returns
        -------
        Beam | None
            Beam with the requested ID. If the ID is not found, ``None`` is returned.

        Notes
        -----
        This method searches for beams in the component and nested components
        recursively.
        """
        # Search in component's beams
        for beam in self.beams:
            if beam.id == id and beam.is_alive:
                return beam

        # If no luck, search on nested components
        result = None
        for component in self.components:
            result = component.search_beam(id)
            if result:
                return result

        # If you reached this point... this means that no beam was found!
        return None

    def _kill_component_on_client(self) -> None:
        """Set the ``is_alive`` property of nested objects to ``False``.

        Notes
        -----
        This method is recursive. It is only to be used by the
        :func:`delete_component()` method and itself.
        """
        # Kill all its bodies, beams and coordinate systems
        for elem in [*self.bodies, *self.beams, *self._coordinate_systems]:
            elem._is_alive = False

        # Now, go to the nested components and kill them as well
        for component in self.components:
            component._kill_component_on_client()

        # Kill itself
        self._is_alive = False

    @graphics_required
    def tessellate(
        self, tess_options: TessellationOptions | None = None, _recursive_call: bool = False
    ) -> Union["PolyData", list["MultiBlock"]]:
        """Tessellate the component.

        Parameters
        ----------
        tess_options : TessellationOptions | None, default: None
            A set of options to determine the tessellation quality.
        _recursive_call: bool, default: False
            Internal flag to indicate if this method is being called recursively.
            Not to be used by the user.

        Returns
        -------
        ~pyvista.PolyData, list[~pyvista.MultiBlock]
            Tessellated component as a single PolyData object.
            If the method is called recursively, a list of MultiBlock objects is returned.

        """
        import pyvista as pv

        # Tessellate the bodies in this component
        datasets: list["MultiBlock"] = [
            body.tessellate(merge=False, tess_options=tess_options) for body in self.bodies
        ]

        # Now, go recursively inside its subcomponents (with no arguments) and
        # merge the PolyData obtained into our blocks
        for comp in self._components:
            if not comp.is_alive:
                continue
            datasets.extend(comp.tessellate(tess_options=tess_options, _recursive_call=True))

        # Convert to polydata as it's slightly faster than extract surface
        # plus this method is only for visualizing the component as a whole (no
        # need to keep the hierarchy)
        if _recursive_call:
            return datasets
        else:
            ugrid = pv.MultiBlock(datasets).combine()
            return pv.PolyData(var_inp=ugrid.points, faces=ugrid.cells)

    @graphics_required
    def plot(
        self,
        merge_component: bool = True,
        merge_bodies: bool = True,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        allow_picking: bool | None = None,
        **plotting_options: dict | None,
    ) -> None | list[Any]:
        """Plot the component.

        Parameters
        ----------
        merge_component : bool, default: True
            Whether to merge the component into a single dataset. By default, ``True``.
            Performance improved. When ``True``, all the faces of the component are effectively
            merged into a single dataset. If ``False``, the individual bodies are kept separate.
        merge_bodies : bool, default: True
            Whether to merge each body into a single dataset. By default, ``True``.
            Performance improved. When ``True``, all the faces of each individual body are
            effectively merged into a single dataset. If ``False``, the individual faces are kept
            separate.
        screenshot : str, default: None
            Path for saving a screenshot of the image being represented.
        use_trame : bool, default: None
            Whether to enable the use of `trame <https://kitware.github.io/trame/index.html>`_.
            The default is ``None``, in which case the
            ``ansys.tools.visualization_interface.USE_TRAME`` global setting is used.
        use_service_colors : bool, default: None
            Whether to use the colors assigned to the body in the service. The default
            is ``None``, in which case the ``ansys.geometry.core.USE_SERVICE_COLORS``
            global setting is used.
        allow_picking : bool, default: None
            Whether to enable picking. The default is ``None``, in which case the
            picker is not enabled.
        **plotting_options : dict, default: None
            Keyword arguments for plotting. For allowable keyword arguments, see the

        Returns
        -------
        None | list[Any]
            If ``allow_picking=True``, a list of picked objects is returned. Otherwise, ``None``.

        Examples
        --------
        Create 25 small cylinders in a grid-like pattern on the XY plane and
        plot them. Make the cylinders look metallic by enabling
        physically-based rendering with ``pbr=True``.

        >>> from ansys.geometry.core.misc.units import UNITS as u
        >>> from ansys.geometry.core.sketch import Sketch
        >>> from ansys.geometry.core.math import Plane, Point2D, Point3D, UnitVector3D
        >>> from ansys.geometry.core import Modeler
        >>> import numpy as np
        >>> modeler = Modeler()
        >>> origin = Point3D([0, 0, 0])
        >>> plane = Plane(origin, direction_x=[1, 0, 0], direction_y=[0, 1, 0])
        >>> design = modeler.create_design("my-design")
        >>> mycomp = design.add_component("my-comp")
        >>> n = 5
        >>> xx, yy = np.meshgrid(
        ...     np.linspace(-4, 4, n),
        ...     np.linspace(-4, 4, n),
        ... )
        >>> for x, y in zip(xx.ravel(), yy.ravel()):
        ...     sketch = Sketch(plane)
        ...     sketch.circle(Point2D([x, y]), 0.2 * u.m)
        ...     mycomp.extrude_sketch(f"body-{x}-{y}", sketch, 1 * u.m)
        >>> mycomp
        ansys.geometry.core.designer.Component 0x2203cc9ec50
            Name                 : my-comp
            Exists               : True
            Parent component     : my-design
            N Bodies             : 25
            N Components         : 0
            N Coordinate Systems : 0
        >>> mycomp.plot(pbr=True, metallic=1.0)
        """
        import ansys.geometry.core as pyansys_geometry
        from ansys.geometry.core.plotting import GeometryPlotter

        use_service_colors = (
            use_service_colors
            if use_service_colors is not None
            else pyansys_geometry.USE_SERVICE_COLORS
        )
        # If picking is enabled, we should not merge the component
        if allow_picking:
            # This blocks the user from selecting the component itself
            # but honestly, who would want to select the component itself since
            # you already have a reference to it? It is the object you are plotting!
            self._grpc_client.log.info(
                "Ignoring 'merge_component=True' (default behavior) as "
                "'allow_picking=True' has been requested."
            )
            merge_component = False

        # Add merge_component and merge_bodies to the plotting options
        plotting_options["merge_component"] = merge_component
        plotting_options["merge_bodies"] = merge_bodies

        # At component level, if ``multi_colors`` or ``use_service_colors`` are defined
        # we should not merge the component or the bodies (only if ``use_service_colors`` is True).
        if plotting_options.get("multi_colors", False) or use_service_colors:
            plotting_options["merge_component"] = False
            self._grpc_client.log.info(
                "Ignoring 'merge_component=True' (default behavior) as "
                "'multi_colors' or 'use_service_colors' are defined."
            )
            if use_service_colors:
                plotting_options["merge_bodies"] = False
                self._grpc_client.log.info(
                    "Ignoring 'merge_bodies=True' (default behavior) as "
                    "'use_service_colors' is defined."
                )

        pl = GeometryPlotter(
            use_trame=use_trame,
            use_service_colors=use_service_colors,
            allow_picking=allow_picking,
        )
        pl.plot(self, **plotting_options)
        return pl.show(screenshot=screenshot, **plotting_options)

    def __repr__(self) -> str:
        """Represent the ``Component`` as a string."""
        alive_bodies = [1 if body.is_alive else 0 for body in self.bodies]
        alive_beams = [1 if beam.is_alive else 0 for beam in self.beams]
        alive_coords = [1 if cs.is_alive else 0 for cs in self.coordinate_systems]
        alive_comps = [1 if comp.is_alive else 0 for comp in self.components]
        lines = [f"ansys.geometry.core.designer.Component {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Parent component     : {self.parent_component.name}")
        lines.append(f"  N Bodies             : {sum(alive_bodies)}")
        lines.append(f"  N Beams              : {sum(alive_beams)}")
        lines.append(f"  N Coordinate Systems : {sum(alive_coords)}")
        lines.append(f"  N Design Points      : {len(self.design_points)}")
        lines.append(f"  N Components         : {sum(alive_comps)}")
        return "\n".join(lines)

    @check_input_types
    def tree_print(
        self,
        consider_comps: bool = True,
        consider_bodies: bool = True,
        consider_beams: bool = True,
        depth: int | None = None,
        indent: int = 4,
        sort_keys: bool = False,
        return_list: bool = False,
        skip_loc_header: bool = False,
    ) -> None | list[str]:
        """Print the component in tree format.

        Parameters
        ----------
        consider_comps : bool, default: True
            Whether to print the nested components.
        consider_bodies : bool, default: True
            Whether to print the bodies.
        consider_beams : bool, default: True
            Whether to print the beams.
        depth : int | None, default: None
            Depth level to print. If None, it prints all levels.
        indent : int, default: 4
            Indentation level. Minimum is 2 - if less than 2, it is set to 2
            by default.
        sort_keys : bool, default: False
            Whether to sort the keys alphabetically.
        return_list : bool, default: False
            Whether to return a list of strings or print out
            the tree structure.
        skip_loc_header : bool, default: False
            Whether to skip the location header. Mostly for internal use.

        Returns
        -------
        None | list[str]
            Tree-style printed component or list of strings representing the component tree.
        """

        def build_parent_tree(comp: Component, parent_tree: str = "") -> str:
            """Private function to build the parent tree of a component."""
            if comp.parent_component is None:
                # We reached the top level component... return the parent tree
                return "Root component (Design)" if not parent_tree else parent_tree
            else:
                if parent_tree == "":
                    # Should only happen in the first call
                    parent_tree = comp.name

                # Add the parent component to the parent tree and continue
                return build_parent_tree(
                    comp.parent_component, f"{comp.parent_component.name} > {parent_tree}"
                )

        # Indentation should be at least 2
        indent = max(2, indent)

        # Initialize the lines list
        lines: list[str] = []

        # Add the location header if requested - and only on the first call
        # (subsequent calls will have the skip_loc_header set to True)
        if not skip_loc_header:
            lines.append(f">>> Tree print view of component '{self.name}'")
            lines.append("")
            lines.append("Location")
            lines.append(f"{'-' * len(lines[-1])}")
            lines.append(f"{build_parent_tree(self)}")
            lines.append("")
            lines.append("Subtree")
            lines.append(f"{'-' * len(lines[-1])}")

        lines.append(f"(comp) {self.name}")
        # Print the bodies
        if consider_bodies:
            # Check if the bodies should be sorted
            if sort_keys:
                body_names = [body.name for body in sorted(self.bodies, key=lambda body: body.name)]
            else:
                body_names = [body.name for body in self.bodies]

            # Add the bodies to the lines (with indentation)
            lines.extend([f"|{'-' * (indent - 1)}(body) {name}" for name in body_names])

        # Print the beams
        if consider_beams:
            # Check if the bodies should be sorted
            if sort_keys:
                # TODO: Beams should also have names...
                # https://github.com/ansys/pyansys-geometry/issues/1319
                beam_names = [
                    beam.id
                    for beam in sorted(self.beams, key=lambda beam: beam.id)
                    if beam.is_alive
                ]
            else:
                beam_names = [beam.id for beam in self.beams if beam.is_alive]

            # Add the bodies to the lines (with indentation)
            lines.extend([f"|{'-' * (indent - 1)}(beam) {name}" for name in beam_names])

        # Print the nested components
        if consider_comps:
            # Check if the components should be sorted
            comps = (
                self.components
                if not sort_keys
                else sorted(self.components, key=lambda comp: comp.name)
            )
            comps = [comp for comp in comps if comp.is_alive]

            # Add the components to the lines (recursive)
            if depth is None or depth > 1:
                n_comps = len(comps)
                for idx, comp in enumerate(comps):
                    subcomp = comp.tree_print(
                        consider_comps=consider_comps,
                        consider_bodies=consider_bodies,
                        consider_beams=consider_beams,
                        depth=None if depth is None else depth - 1,
                        indent=indent,
                        sort_keys=sort_keys,
                        return_list=True,
                        skip_loc_header=True,
                    )

                    # Add indentation to the subcomponent lines
                    lines.append(f"|{'-' * (indent - 1)}(comp) {comp.name}")

                    # Determine the prefix for the subcomponent lines and add them
                    prefix = f"{' ' * indent}" if idx == (n_comps - 1) else f":{' ' * (indent - 1)}"
                    lines.extend([f"{prefix}{line}" for line in subcomp[1:]])

            else:
                lines.extend([f"|{'-' * (indent - 1)}(comp) {comp.name}" for comp in comps])

        return lines if return_list else print("\n".join(lines))



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing 2D and 3D vectors."""

from io import UnsupportedOperation
from typing import Union

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity

from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.misc.accuracy import Accuracy
from ansys.geometry.core.misc.checks import check_ndarray_is_float_int
from ansys.geometry.core.misc.measurements import Angle
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.typing import Real, RealSequence

_NUMPY_MAJOR_VERSION = int(np.__version__[0])
"""Major version of the numpy library."""


class Vector3D(np.ndarray):
    """Provides for managing and creating a 3D vector.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence
        3D :class:`numpy.ndarray <numpy.ndarray>` class with shape(X,).
    """

    def __new__(cls, input: np.ndarray | RealSequence):
        """Initialize the ``Vector3D`` class."""
        obj = np.asarray(input).view(cls)

        # Check that the size is as expected
        if len(obj) != 3:
            raise ValueError("Vector3D class must receive 3 arguments.")  # noqa: E501

        # Check the input data
        check_ndarray_is_float_int(obj, "input")

        return obj

    @property
    def x(self) -> Real:
        """X coordinate of the ``Vector3D`` class."""
        return self[0]

    @x.setter
    @check_input_types
    def x(self, value: Real) -> None:
        """Set the Y coordinate of the ``Vector3D`` class."""
        self[0] = value

    @property
    def y(self) -> Real:
        """Y coordinate of the ``Vector3D`` class."""
        return self[1]

    @y.setter
    @check_input_types
    def y(self, value: Real) -> None:
        """Set the Y coordinate of the ``Vector3D`` class."""
        self[1] = value

    @property
    def z(self) -> Real:
        """Z coordinate of the ``Vector3D`` class."""
        return self[2]

    @z.setter
    @check_input_types
    def z(self, value: Real) -> None:
        """Set the Z coordinate of the ``Vector3D`` class."""
        self[2] = value

    @property
    def norm(self) -> float:
        """Norm of the vector."""
        return np.linalg.norm(self)

    @property
    def magnitude(self) -> float:
        """Norm of the vector."""
        return self.norm

    @property
    def is_zero(self) -> bool:
        """Check if all components of the 3D vector are zero."""
        return all([comp == 0 for comp in self])

    @check_input_types
    def is_perpendicular_to(self, other_vector: "Vector3D") -> bool:
        """Check if this vector and another vector are perpendicular."""
        if self.is_zero or other_vector.is_zero:
            return False
        else:
            return Accuracy.angle_is_zero(self * other_vector)

    @check_input_types
    def is_parallel_to(self, other_vector: "Vector3D") -> bool:
        """Check if this vector and another vector are parallel."""
        if self.is_zero or other_vector.is_zero:
            return False
        else:
            return (self % other_vector).is_zero

    @check_input_types
    def is_opposite(self, other_vector: "Vector3D") -> bool:
        """Check if this vector and another vector are opposite."""
        return bool(self.is_parallel_to(other_vector) and self * other_vector < 0)

    def normalize(self) -> "Vector3D":
        """Return a normalized version of the 3D vector."""
        norm = self.norm
        if norm > 0:
            return (self / norm).view(Vector3D)
        else:
            raise ValueError("The norm of the 3D vector is not valid.")

    def transform(self, matrix: "Matrix44") -> "Vector3D":
        """Transform the 3D vector3D with a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4x4 transformation matrix to apply to the vector.

        Returns
        -------
        Vector3D
            A new 3D vector that is the transformed copy of the original 3D vector after applying
            the transformation matrix.

        Notes
        -----
        Transform the ``Vector3D`` object by applying the specified 4x4
        transformation matrix and return a new ``Vector3D`` object representing the
        transformed vector.
        """
        vector_4x1 = np.append(self, 1)
        result_4x1 = matrix * vector_4x1
        result_vector = Vector3D(result_4x1[0:3])
        return result_vector

    @check_input_types
    def rotate_vector(self, vector: "Vector3D", angle: Real | Quantity | Angle) -> "Vector3D":
        """Rotate a vector around a given axis by a specified angle."""
        if self.is_zero:
            raise Exception("Invalid vector operation: rotation axis cannot be zero.")

        # Convert angle to Angle object and get its value in radians
        angle = angle if isinstance(angle, Angle) else Angle(angle)
        angle_m = angle.value.m_as(UNITS.radian)

        axis = self.normalize()
        parallel = axis * (vector.dot(axis))
        perpendicular1 = vector - parallel
        perpendicular2 = axis.cross(perpendicular1)
        return parallel + perpendicular1 * np.cos(angle_m) + perpendicular2 * np.sin(angle_m)

    @check_input_types
    def get_angle_between(self, v: "Vector3D") -> Quantity:
        """Get the angle between this 3D vector and another 3D vector.

        Parameters
        ----------
        v : Vector3D
            Other 3D vector for computing the angle.

        Returns
        -------
        ~pint.Quantity
            Angle between these two 3D vectors.
        """
        if v.is_zero or self.is_zero:
            raise ValueError("Vectors cannot be zero-valued.")

        sine = (self % v).magnitude
        cosine = self * v

        if Accuracy.angle_is_zero(sine):
            if cosine > 0.0:
                return Quantity(0, UNITS.radian)
            else:
                return Quantity(np.pi, UNITS.radian)
        else:
            return Quantity(np.arctan2(sine, cosine), UNITS.radian)

    @check_input_types
    def cross(self, v: "Vector3D") -> "Vector3D":
        """Return the cross product of ``Vector3D`` objects."""
        return np.cross(self, v).view(Vector3D)

    @check_input_types
    def __eq__(self, other: "Vector3D") -> bool:
        """Equals operator for the ``Vector3D`` class."""
        return np.array_equal(self, other)

    def __ne__(self, other: "Vector3D") -> bool:
        """Not equals operator for the ``Vector3D`` class."""
        return not self == other

    @check_input_types
    def __mul__(self, other: Union["Vector3D", Real]) -> Union["Vector3D", Real]:
        """Overload * operator with dot product.

        Notes
        -----
        This method also admits scalar multiplication.
        """
        if isinstance(other, (int, float)):
            return np.multiply(self, other).view(Vector3D)
        else:
            return self.dot(other)

    def __mod__(self, other: "Vector3D") -> "Vector3D":
        """Overload % operator with cross product."""
        return self.cross(other)

    @check_input_types
    def __add__(self, other: Union["Vector3D", Point3D]) -> Union["Vector3D", Point3D]:
        """Addition operation overload for 3D vectors."""
        if isinstance(other, Point3D):
            return other + self
        else:
            return Vector3D(np.add(self, other))

    @check_input_types
    def __sub__(self, other: "Vector3D") -> "Vector3D":
        """Subtraction operation overload for 3D vectors."""
        return np.subtract(self, other).view(Vector3D)

    @classmethod
    @check_input_types
    def from_points(
        cls,
        point_a: np.ndarray | RealSequence | Point3D,
        point_b: np.ndarray | RealSequence | Point3D,
    ):
        """Create a 3D vector from two distinct 3D points.

        Parameters
        ----------
        point_a : ~numpy.ndarray | RealSequence | Point3D
            :class:`Point3D <ansys.geometry.core.math.point.Point3D>`
            class representing the first point.
        point_b : ~numpy.ndarray | RealSequence | Point3D
            :class:`Point3D <ansys.geometry.core.math.point.Point3D>`
            class representing the second point.

        Returns
        -------
        Vector3D
            3D vector from ``point_a`` to ``point_b``.

        Notes
        -----
        The resulting 3D vector is always expressed in ``Point3D``
        base units.
        """
        return Vector3D(point_b - point_a)


class Vector2D(np.ndarray):
    """Provides for creating and managing a 2D vector.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence
        2D :class:`numpy.ndarray <numpy.ndarray>` class with shape(X,).
    """

    def __new__(cls, input: np.ndarray | RealSequence):
        """Initialize the ``Vector2D`` class."""
        obj = np.asarray(input).view(cls)

        # Check that the size is as expected
        if len(obj) != 2:
            raise ValueError("Vector2D class must receive 2 arguments.")  # noqa: E501

        # Check the input data
        check_ndarray_is_float_int(obj, "input")

        return obj

    @property
    def x(self) -> Real:
        """X coordinate of the 2D vector."""
        return self[0]

    @x.setter
    @check_input_types
    def x(self, value: Real) -> None:
        """Set the X coordinate of the 2D vector."""
        self[0] = value

    @property
    def y(self) -> Real:
        """Y coordinate of the 2D vector."""
        return self[1]

    @y.setter
    @check_input_types
    def y(self, value: Real) -> None:
        """Set the Y coordinate of the 2D vector."""
        self[1] = value

    @property
    def norm(self) -> float:
        """Norm of the 2D vector."""
        return np.linalg.norm(self)

    @property
    def magnitude(self) -> float:
        """Norm of the 2D vector."""
        return self.norm

    @property
    def is_zero(self) -> bool:
        """Check if values for all components of the 2D vector are zero."""
        return all([comp == 0 for comp in self])

    @check_input_types
    def cross(self, v: "Vector2D") -> Real:
        """Return the cross product of ``Vector2D`` objects."""
        if _NUMPY_MAJOR_VERSION >= 2:
            # See https://github.com/numpy/numpy/issues/26620 and more specifically
            # https://github.com/numpy/numpy/issues/26620#issuecomment-2150748569
            return self[..., 0] * v[..., 1] - self[..., 1] * v[..., 0]
        else:  # pragma: no cover
            # Coverage is measured with the latest version of numpy
            # so this code is not covered
            return np.cross(self, v)

    @check_input_types
    def is_perpendicular_to(self, other_vector: "Vector2D") -> bool:
        """Check if this 2D vector and another 2D vector are perpendicular."""
        if self.is_zero or other_vector.is_zero:
            return False
        else:
            return Accuracy.angle_is_zero(self * other_vector)

    @check_input_types
    def is_parallel_to(self, other_vector: "Vector2D") -> bool:
        """Check if this vector and another vector are parallel."""
        if self.is_zero or other_vector.is_zero:
            return False
        else:
            return bool((self % other_vector) == 0)

    @check_input_types
    def is_opposite(self, other_vector: "Vector2D") -> bool:
        """Check if this vector and another vector are opposite."""
        return bool(self.is_parallel_to(other_vector) and self * other_vector < 0)

    def normalize(self) -> "Vector2D":
        """Return a normalized version of the 2D vector."""
        norm = self.norm
        if norm > 0:
            return (self / norm).view(Vector2D)
        else:
            raise ValueError("The norm of the 2D vector is not valid.")

    @check_input_types
    def get_angle_between(self, v: "Vector2D") -> Quantity:
        """Get the angle between this 2D vector and another 2D vector.

        Parameters
        ----------
        v : Vector2D
            Other 2D vector to compute the angle with.

        Returns
        -------
        ~pint.Quantity
            Angle between these two 2D vectors.
        """
        if v.is_zero or self.is_zero:
            raise ValueError("Vectors cannot be zero-valued.")

        angle = np.arctan2(v.y * self.x - v.x * self.y, v.x * self.x + v.y * self.y)

        if angle < 0:
            angle = angle + 2 * np.pi

        return Quantity(angle, UNITS.radian)

    @check_input_types
    def __eq__(self, other: "Vector2D") -> bool:
        """Equals operator for the ``Vector2D`` class."""
        return np.array_equal(self, other)

    def __ne__(self, other: "Vector2D") -> bool:
        """Not equals operator for the ``Vector2D`` class."""
        return not self == other

    @check_input_types
    def __mul__(self, other: Union["Vector2D", Real]) -> Union["Vector2D", Real]:
        """Overload * operator with dot product.

        Notes
        -----
        This method also admits scalar multiplication.
        """
        if isinstance(other, (int, float)):
            return np.multiply(self, other).view(Vector2D)
        else:
            return self.dot(other)

    @check_input_types
    def __add__(self, other: Union["Vector2D", Point2D]) -> Union["Vector2D", Point2D]:
        """Addition operation overload for 2D vectors."""
        if isinstance(other, Point2D):
            return other + self
        else:
            return Vector2D(np.add(self, other))

    @check_input_types
    def __sub__(self, other: "Vector2D") -> "Vector2D":
        """Subtraction operation overload for 2D vectors."""
        return np.subtract(self, other).view(Vector2D)

    def __mod__(self, other: "Vector2D") -> Real:
        """Overload % operator with cross product."""
        return self.cross(other)

    @classmethod
    @check_input_types
    def from_points(
        cls,
        point_a: np.ndarray | RealSequence | Point2D,
        point_b: np.ndarray | RealSequence | Point2D,
    ):
        """Create a 2D vector from two distinct 2D points.

        Parameters
        ----------
        point_a : ~numpy.ndarray | RealSequence | Point2D
            :class:`Point2D <ansys.geometry.core.math.point.Point2D>`
            class representing the first point.
        point_b : ~numpy.ndarray | RealSequence | Point2D
            :class:`Point2D <ansys.geometry.core.math.point.Point2D>`
            class representing the second point.

        Returns
        -------
        Vector2D
            2D vector from ``point_a`` to ``point_b``.

        Notes
        -----
        The resulting 2D vector is always expressed in ``Point2D``
        base units.
        """
        return Vector2D(point_b - point_a)


class UnitVector3D(Vector3D):
    """Provides for creating and managing a 3D unit vector.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence | Vector3D
        * 1D :class:`numpy.ndarray <numpy.ndarray>` class with shape(X,)
        * Vector3D
    """

    def __new__(cls, input: np.ndarray | RealSequence | Vector3D):
        """Initialize the ``UnitVector3D`` class."""
        obj = Vector3D(input) if not isinstance(input, Vector3D) else input
        obj = obj.normalize().view(cls)
        obj.setflags(write=False)
        return obj

    @Vector3D.x.setter
    def x(self, value: Real) -> None:  # noqa : D102
        raise UnsupportedOperation("UnitVector3D is immutable.")

    @Vector3D.y.setter
    def y(self, value: Real) -> None:  # noqa : D102
        raise UnsupportedOperation("UnitVector3D is immutable.")

    @Vector3D.z.setter
    def z(self, value: Real) -> None:  # noqa : D102
        raise UnsupportedOperation("UnitVector3D is immutable.")

    @classmethod
    def from_points(
        cls,
        point_a: np.ndarray | RealSequence | Point3D,
        point_b: np.ndarray | RealSequence | Point3D,
    ):
        """Create a 3D unit vector from two distinct 3D points.

        Parameters
        ----------
        point_a : ~numpy.ndarray | RealSequence | Point3D
            :class:`Point3D <ansys.geometry.core.math.point.Point3D>`
            class representing the first point.
        point_b : ~numpy.ndarray | RealSequence | Point3D
            :class:`Point3D <ansys.geometry.core.math.point.Point3D>`
            class representing the second point.

        Returns
        -------
        UnitVector3D
            3D unit vector from ``point_a`` to ``point_b``.
        """
        return UnitVector3D(Vector3D.from_points(point_a, point_b))


class UnitVector2D(Vector2D):
    """Provides for creating and managing a 3D unit vector.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence | Vector2D
        * 1D :class:`numpy.ndarray <numpy.ndarray>` class with shape(X,)
        * Vector2D
    """

    def __new__(cls, input: np.ndarray | RealSequence | Vector2D):
        """Initialize the ``UnitVector2D`` class."""
        obj = Vector2D(input) if not isinstance(input, Vector2D) else input
        obj = obj.normalize().view(cls)
        obj.setflags(write=False)
        return obj

    @Vector2D.x.setter
    def x(self, value: Real) -> None:  # noqa : D102
        raise UnsupportedOperation("UnitVector2D is immutable.")

    @Vector2D.y.setter
    def y(self, value: Real) -> None:  # noqa : D102
        raise UnsupportedOperation("UnitVector2D is immutable.")

    @classmethod
    def from_points(
        cls,
        point_a: np.ndarray | RealSequence | Point2D,
        point_b: np.ndarray | RealSequence | Point2D,
    ):
        """Create a 2D unit vector from two distinct 2D points.

        Parameters
        ----------
        point_a : ~numpy.ndarray | RealSequence | Point2D
            :class:`Point2D <ansys.geometry.core.math.point.Point2D>`
            class representing the first point.
        point_b : ~numpy.ndarray | RealSequence | Point2D
            :class:`Point2D <ansys.geometry.core.math.point.Point2D>`
            class representing the second point.

        Returns
        -------
        UnitVector2D
            2D unit vector from ``point_a`` to ``point_b``.
        """
        return UnitVector2D(Vector2D.from_points(point_a, point_b))



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides matrix primitive representations."""

from typing import TYPE_CHECKING, Union

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.misc.checks import check_ndarray_is_float_int, check_type
from ansys.geometry.core.typing import Real, RealSequence

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.math.frame import Frame
    from ansys.geometry.core.math.vector import Vector3D  # For type hints

DEFAULT_MATRIX33 = np.identity(3)
"""Default value of the 3x3 identity matrix for the ``Matrix33`` class."""

DEFAULT_MATRIX44 = np.identity(4)
"""Default value of the 4x4 identity matrix for the ``Matrix44`` class."""


class Matrix(np.ndarray):
    """Provides matrix representation.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence
        Matrix arguments as a :class:`np.ndarray <numpy.ndarray>` class.
    """

    def __new__(cls, input: np.ndarray | RealSequence):
        """Initialize ``Matrix`` class."""
        obj = np.asarray(input).view(cls)
        obj.setflags(write=False)

        if obj is None or obj.ndim != 2:
            raise ValueError("Matrix should only be a 2D array.")

        check_ndarray_is_float_int(obj)

        return obj

    def determinant(self) -> Real:
        """Get the determinant of the matrix."""
        if self.shape[0] != self.shape[1]:
            raise ValueError("The determinant is only defined for square matrices.")
        return np.linalg.det(self)

    def inverse(self) -> "Matrix":
        """Provide the inverse of the matrix."""
        det = self.determinant()
        if det <= 0:
            raise ValueError("The matrix cannot be inversed because its determinant is zero.")
        return np.linalg.inv(self)

    @check_input_types
    def __mul__(self, other: Union["Matrix", np.ndarray]) -> "Matrix":
        """Get the multiplication of the matrix."""
        if self.shape[1] != other.shape[0]:
            raise ValueError(
                f"The dimensions of the matrices {self.shape[1]} and {other.shape[0]} are not multipliable."  # noqa : E501
            )
        return np.matmul(self, other).view(Matrix)

    @check_input_types
    def __eq__(self, other: "Matrix") -> bool:
        """Equals operator for the ``Matrix`` class."""
        return np.array_equal(self, other)

    def __ne__(self, other: "Matrix") -> bool:
        """Not equals operator for the ``Matrix`` class."""
        return not self == other


class Matrix33(Matrix):
    """Provides 3x3 matrix representation.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence | Matrix, default: DEFAULT_MATRIX33
        Matrix arguments as a :class:`np.ndarray <numpy.ndarray>` class.
    """

    def __new__(cls, input: np.ndarray | RealSequence | Matrix = DEFAULT_MATRIX33):
        """Initialize the ``Matrix33`` class."""
        obj = Matrix(input).view(cls)
        if input is DEFAULT_MATRIX33:
            return obj

        if obj.shape != (3, 3):
            raise ValueError("Matrix33 should only be a 2D array of shape (3,3).")

        return obj


class Matrix44(Matrix):
    """Provides 4x4 matrix representation.

    Parameters
    ----------
    input : ~numpy.ndarray | RealSequence | Matrix, default: DEFAULT_MATRIX44
        Matrix arguments as a :class:`np.ndarray <numpy.ndarray>` class.
    """

    def __new__(cls, input: np.ndarray | RealSequence | Matrix = DEFAULT_MATRIX44):
        """Initialize the ``Matrix44`` class."""
        obj = Matrix(input).view(cls)
        if input is DEFAULT_MATRIX44:
            return obj

        if obj.shape != (4, 4):
            raise ValueError("Matrix44 should only be a 2D array of shape (4,4).")

        return obj

    @classmethod
    def create_translation(cls, translation: "Vector3D") -> "Matrix44":
        """Create a matrix representing the specified translation.

        Parameters
        ----------
        translation : Vector3D
            The translation vector representing the translation. The components of the vector
            should be in meters.

        Returns
        -------
        Matrix44
            A 4x4 matrix representing the translation.

        Examples
        --------
        >>> translation_vector = Vector3D(1.0, 2.0, 3.0)
        >>> translation_matrix = Matrix44.create_translation(translation_vector)
        >>> print(translation_matrix)
        [[1. 0. 0. 1.]
         [0. 1. 0. 2.]
         [0. 0. 1. 3.]
         [0. 0. 0. 1.]]
        """
        from ansys.geometry.core.math.vector import Vector3D

        # Verify the input
        check_type(translation, Vector3D)

        matrix = cls(
            [
                [1, 0, 0, translation.x],
                [0, 1, 0, translation.y],
                [0, 0, 1, translation.z],
                [0, 0, 0, 1],
            ]
        )
        return matrix

    def is_translation(self, including_identity: bool = False) -> bool:
        """Check if the matrix represents a translation.

        This method checks if the matrix represents a translation transformation.
        A translation matrix has the following form:

            [1 0 0 tx]
            [0 1 0 ty]
            [0 0 1 tz]
            [0 0 0  1]

        Parameters
        ----------
        including_identity : bool, optional
            If ``True``, the method will return ``True`` for the identity matrix as well.
            If ``False``, the method will return ``False`` for the identity matrix.

        Returns
        -------
        bool
            ``True`` if the matrix represents a translation, ``False`` otherwise.

        Examples
        --------
        >>> matrix = Matrix44([[1, 0, 0, 5], [0, 1, 0, 3], [0, 0, 1, 2], [0, 0, 0, 1]])
        >>> matrix.is_translation()
        True
        >>> identity_matrix = Matrix44([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
        >>> identity_matrix.is_translation()
        False
        >>> identity_matrix.is_translation(including_identity=True)
        True
        """
        if not (
            self.__is_close(self[0][0], 1)
            and self.__is_close(self[0][1], 0)
            and self.__is_close(self[0][2], 0)
        ):
            return False
        if not (
            self.__is_close(self[1][0], 0)
            and self.__is_close(self[1][1], 1)
            and self.__is_close(self[1][2], 0)
        ):
            return False
        if not (
            self.__is_close(self[2][0], 0)
            and self.__is_close(self[2][1], 0)
            and self.__is_close(self[2][2], 1)
        ):
            return False
        if not self.__is_close(self[2][2], 1):
            return False

        if (
            not including_identity
            and self.__is_close(self[0][3], 0)
            and self.__is_close(self[1][3], 0)
            and self.__is_close(self[2][3], 0)
        ):
            return False

        return True

    def __is_close(self, a, b, tol=1e-9):
        """Check if two values are close to each other within a tolerance."""
        return np.isclose(a, b, atol=tol)

    @classmethod
    def create_rotation(
        cls, direction_x: "Vector3D", direction_y: "Vector3D", direction_z: "Vector3D" = None
    ) -> "Matrix44":
        """Create a matrix representing the specified rotation.

        Parameters
        ----------
        direction_x : Vector3D
            The X direction vector.
        direction_y : Vector3D
            The Y direction vector.
        direction_z : Vector3D, optional
            The Z direction vector. If not provided, it will be calculated
            as the cross product of direction_x and direction_y.

        Returns
        -------
        Matrix44
            A 4x4 matrix representing the rotation.

        Examples
        --------
        >>> direction_x = Vector3D(1.0, 0.0, 0.0)
        >>> direction_y = Vector3D(0.0, 1.0, 0.0)
        >>> rotation_matrix = Matrix44.create_rotation(direction_x, direction_y)
        >>> print(rotation_matrix)
        [[1. 0. 0. 0.]
        [0. 1. 0. 0.]
        [0. 0. 1. 0.]
        [0. 0. 0. 1.]]
        """
        from ansys.geometry.core.math.vector import Vector3D

        # Verify the inputs
        check_type(direction_x, Vector3D)
        check_type(direction_y, Vector3D)
        if direction_z is not None:
            check_type(direction_z, Vector3D)

        if not direction_x.is_perpendicular_to(direction_y):
            raise ValueError("The provided direction vectors are not orthogonal.")

        # Normalize the vectors
        direction_x = direction_x.normalize()
        direction_y = direction_y.normalize()

        # Calculate the third direction vector if not provided
        if direction_z is None:
            direction_z = direction_x.cross(direction_y)
        else:
            if not (
                direction_x.is_perpendicular_to(direction_z)
                and direction_y.is_perpendicular_to(direction_z)
            ):
                raise ValueError("The provided direction vectors are not orthogonal.")
            direction_z = direction_z.normalize()

        matrix = cls(
            [
                [direction_x.x, direction_y.x, direction_z.x, 0],
                [direction_x.y, direction_y.y, direction_z.y, 0],
                [direction_x.z, direction_y.z, direction_z.z, 0],
                [0, 0, 0, 1],
            ]
        )
        return matrix

    @classmethod
    def create_matrix_from_rotation_about_axis(cls, axis: "Vector3D", angle: float) -> "Matrix44":
        """
        Create a matrix representing a rotation about a given axis.

        Parameters
        ----------
        axis : Vector3D
            The axis of rotation.
        angle : float
            The angle of rotation in radians.

        Returns
        -------
        Matrix44
            A 4x4 matrix representing the rotation.
        """
        axis_dir = axis.normalize()
        x, y, z = axis_dir[0], axis_dir[1], axis_dir[2]

        k = np.array([[0, -z, y], [z, 0, -x], [-y, x, 0]])

        identity = np.eye(3)
        cos_theta = np.cos(angle)
        sin_theta = np.sin(angle)

        # Rodrigues' rotation formula
        rotation_3x3 = identity + sin_theta * k + (1 - cos_theta) * (k @ k)

        # Convert to a 4x4 homogeneous matrix
        rotation_matrix = np.eye(4)
        rotation_matrix[:3, :3] = rotation_3x3

        return cls(rotation_matrix)

    @classmethod
    def create_matrix_from_mapping(cls, frame: "Frame") -> "Matrix44":
        """
        Create a matrix representing the specified mapping.

        Parameters
        ----------
        frame : Frame
            The frame containing the origin and direction vectors.

        Returns
        -------
        Matrix44
            A 4x4 matrix representing the translation and rotation defined by the frame.
        """
        from ansys.geometry.core.math.vector import Vector3D

        translation_matrix = Matrix44.create_translation(
            Vector3D([frame.origin[0], frame.origin[1], frame.origin[2]])
        )
        rotation_matrix = Matrix44.create_rotation(frame.direction_x, frame.direction_y)
        return translation_matrix * rotation_matrix



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the bodies service implementation (abstraction layer)."""

from abc import ABC, abstractmethod

import grpc


class GRPCBodyService(ABC):  # pragma: no cover
    """Body service for gRPC communication with the Geometry server.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    def __init__(self, channel: grpc.Channel):
        """Initialize the GRPCBodyService class."""
        pass

    @abstractmethod
    def create_sphere_body(self, **kwargs) -> dict:
        """Create a sphere body."""
        pass

    @abstractmethod
    def create_extruded_body(self, **kwargs) -> dict:
        """Create an extruded body."""
        pass

    @abstractmethod
    def create_sweeping_profile_body(self, **kwargs) -> dict:
        """Create a sweeping profile body."""
        pass

    @abstractmethod
    def create_sweeping_chain(self, **kwargs) -> dict:
        """Create a sweeping chain."""
        pass

    @abstractmethod
    def create_extruded_body_from_face_profile(self, **kwargs) -> dict:
        """Create an extruded body from a face profile."""
        pass

    @abstractmethod
    def create_extruded_body_from_loft_profiles(self, **kwargs) -> dict:
        """Create an extruded body from loft profiles."""
        pass

    @abstractmethod
    def create_planar_body(self, **kwargs) -> dict:
        """Create a planar body."""
        pass

    @abstractmethod
    def create_body_from_face(self, **kwargs) -> dict:
        """Create a body from a face."""
        pass

    @abstractmethod
    def create_surface_body(self, **kwargs) -> dict:
        """Create a surface body."""
        pass

    @abstractmethod
    def create_surface_body_from_trimmed_curves(self, **kwargs) -> dict:
        """Create a surface body from trimmed curves."""
        pass

    @abstractmethod
    def translate(self, **kwargs) -> dict:
        """Translate a body."""
        pass

    @abstractmethod
    def delete(self, **kwargs) -> dict:
        """Delete a body."""
        pass

    @abstractmethod
    def is_suppressed(self, **kwargs) -> dict:
        """Check if a body is suppressed."""
        pass

    @abstractmethod
    def get_color(self, **kwargs) -> dict:
        """Get the color of a body."""
        pass

    @abstractmethod
    def get_faces(self, **kwargs) -> dict:
        """Get the faces of a body."""
        pass

    @abstractmethod
    def get_edges(self, **kwargs) -> dict:
        """Get the edges of a body."""
        pass

    @abstractmethod
    def get_volume(self, **kwargs) -> dict:
        """Get the volume of a body."""
        pass

    @abstractmethod
    def get_bounding_box(self, **kwargs) -> dict:
        """Get the bounding box of a body."""
        pass

    @abstractmethod
    def set_assigned_material(self, **kwargs) -> dict:
        """Set the assigned material of a body."""
        pass

    @abstractmethod
    def get_assigned_material(self, **kwargs) -> dict:
        """Get the assigned material of a body."""
        pass

    @abstractmethod
    def set_name(self, **kwargs) -> dict:
        """Set the name of a body."""
        pass

    @abstractmethod
    def set_fill_style(self, **kwargs) -> dict:
        """Set the fill style of a body."""
        pass

    @abstractmethod
    def set_suppressed(self, **kwargs) -> dict:
        """Set the suppressed state of a body."""
        pass

    @abstractmethod
    def set_color(self, **kwargs) -> dict:
        """Set the color of a body."""
        pass

    @abstractmethod
    def rotate(self, **kwargs) -> dict:
        """Rotate a body."""
        pass

    @abstractmethod
    def scale(self, **kwargs) -> dict:
        """Scale a body."""
        pass

    @abstractmethod
    def mirror(self, **kwargs) -> dict:
        """Mirror a body."""
        pass

    @abstractmethod
    def map(self, **kwargs) -> dict:
        """Map a body."""
        pass

    @abstractmethod
    def get_collision(self, **kwargs) -> dict:
        """Get the collision of a body."""
        pass

    @abstractmethod
    def copy(self, **kwargs) -> dict:
        """Copy a body."""
        pass

    @abstractmethod
    def get_tesellation(self, **kwargs) -> dict:
        """Get the tessellation of a body."""
        pass

    @abstractmethod
    def get_tesellation_with_options(self, **kwargs) -> dict:
        """Get the tessellation of a body with options."""
        pass

    @abstractmethod
    def boolean(self, **kwargs) -> dict:
        """Boolean operation."""
        pass



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a user-defined coordinate system."""

from typing import TYPE_CHECKING

from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.math.frame import Frame

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.component import Component


class CoordinateSystem:
    """Represents a user-defined coordinate system within the design assembly.

    This class synchronizes to a design within a supporting Geometry
    service instance.

    Parameters
    ----------
    name : str
        User-defined label for the coordinate system.
    frame : Frame
        Frame defining the coordinate system bounds.
    parent_component : Component, default: Component
        Parent component the coordinate system is assigned against.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    """

    def __init__(
        self,
        name: str,
        frame: Frame,
        parent_component: "Component",
        grpc_client: GrpcClient,
        preexisting_id: str | None = None,
    ):
        """Initialize the ``CoordinateSystem`` class."""
        self._parent_component = parent_component
        self._grpc_client = grpc_client
        self._is_alive = True

        # Create without going to server
        if preexisting_id:
            self._name = name
            self._frame = frame
            self._id = preexisting_id
            return

        self._grpc_client.log.debug("Requesting creation of a coordinate system.")
        response = self._grpc_client.services.coordinate_systems.create(
            parent_id=self._parent_component.id,
            name=name,
            frame=frame,
        )

        self._id = response.get("id")
        self._name = response.get("name")
        self._frame = response.get("frame")

    @property
    def id(self) -> str:
        """ID of the coordinate system."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the coordinate system."""
        return self._name

    @property
    def frame(self) -> Frame:
        """Frame of the coordinate system."""
        return self._frame

    @property
    def parent_component(self) -> "Component":
        """Parent component of the coordinate system."""
        return self._parent_component

    @property
    def is_alive(self) -> bool:
        """Flag indicating if coordinate system is still alive on the server."""
        return self._is_alive

    def __repr__(self) -> str:
        """Represent the coordinate system as a string."""
        lines = [f"ansys.geometry.core.designer.CoordinateSystem {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Parent component     : {self.parent_component.name}")
        lines.append(
            f"  Frame origin         : [{','.join([str(x) for x in self.frame.origin])}] in meters"
        )
        lines.append(
            f"  Frame X-direction    : [{','.join([str(x) for x in self.frame.direction_x])}]"
        )
        lines.append(
            f"  Frame Y-direction    : [{','.join([str(x) for x in self.frame.direction_y])}]"
        )
        lines.append(
            f"  Frame Z-direction    : [{','.join([str(x) for x in self.frame.direction_z])}]"
        )
        return "\n".join(lines)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for managing a face."""

from enum import Enum, unique
from typing import TYPE_CHECKING

from beartype import beartype as check_input_types
import matplotlib.colors as mcolors
from pint import Quantity

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.commands_pb2 import FaceOffsetRequest
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.designer.edge import Edge
from ansys.geometry.core.errors import GeometryRuntimeError, protect_grpc
from ansys.geometry.core.math.bbox import BoundingBox
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.auxiliary import (
    DEFAULT_COLOR,
    convert_color_to_hex,
    convert_opacity_to_hex,
)
from ansys.geometry.core.misc.checks import (
    ensure_design_is_active,
    graphics_required,
    min_backend_version,
)
from ansys.geometry.core.misc.options import TessellationOptions
from ansys.geometry.core.shapes.box_uv import BoxUV
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.shapes.parameterization import Interval
from ansys.geometry.core.shapes.surfaces.trimmed_surface import (
    ReversedTrimmedSurface,
    TrimmedSurface,
)

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv

    from ansys.geometry.core.designer.body import Body


@unique
class SurfaceType(Enum):
    """Provides values for the surface types supported."""

    SURFACETYPE_UNKNOWN = 0
    SURFACETYPE_PLANE = 1
    SURFACETYPE_CYLINDER = 2
    SURFACETYPE_CONE = 3
    SURFACETYPE_TORUS = 4
    SURFACETYPE_SPHERE = 5
    SURFACETYPE_NURBS = 6
    SURFACETYPE_PROCEDURAL = 7


@unique
class FaceLoopType(Enum):
    """Provides values for the face loop types supported."""

    INNER_LOOP = "INNER"
    OUTER_LOOP = "OUTER"


class FaceLoop:
    """Provides an internal class holding the face loops defined.

    Parameters
    ----------
    type : FaceLoopType
        Type of loop.
    length : Quantity
        Length of the loop.
    min_bbox : Point3D
        Minimum point of the bounding box containing the loop.
    max_bbox : Point3D
        Maximum point of the bounding box containing the loop.
    edges : list[Edge]
        Edges contained in the loop.

    Notes
    -----
    This class is to be used only when parsing server side results. It is not
    intended to be instantiated by a user.
    """

    def __init__(
        self,
        type: FaceLoopType,
        length: Quantity,
        min_bbox: Point3D,
        max_bbox: Point3D,
        edges: list[Edge],
    ):
        """Initialize ``FaceLoop`` class."""
        self._type = type
        self._length = length
        self._min_bbox = min_bbox
        self._max_bbox = max_bbox
        self._edges = edges

    @property
    def type(self) -> FaceLoopType:
        """Type of the loop."""
        return self._type

    @property
    def length(self) -> Quantity:
        """Length of the loop."""
        return self._length

    @property
    def min_bbox(self) -> Point3D:
        """Minimum point of the bounding box containing the loop."""
        return self._min_bbox

    @property
    def max_bbox(self) -> Point3D:
        """Maximum point of the bounding box containing the loop."""
        return self._max_bbox

    @property
    def edges(self) -> list[Edge]:
        """Edges contained in the loop."""
        return self._edges


class Face:
    """Represents a single face of a body within the design assembly.

    This class synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    surface_type : SurfaceType
        Type of surface that the face forms.
    body : Body
        Parent body that the face constructs.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    """

    def __init__(
        self,
        id: str,
        surface_type: SurfaceType,
        body: "Body",
        grpc_client: GrpcClient,
        is_reversed: bool = False,
    ):
        """Initialize the ``Face`` class."""
        self._id = id
        self._surface_type = surface_type
        self._body = body
        self._grpc_client = grpc_client
        self._commands_stub = CommandsStub(grpc_client.channel)
        self._is_reversed = is_reversed
        self._shape = None
        self._color = None

    @property
    def id(self) -> str:
        """Face ID."""
        return self._id

    @property
    def _grpc_id(self) -> EntityIdentifier:
        """Entity ID of this face on the server side."""
        return EntityIdentifier(id=self._id)

    @property
    def is_reversed(self) -> bool:
        """Flag indicating if the face is reversed."""
        return self._is_reversed

    @property
    def body(self) -> "Body":
        """Body that the face belongs to."""
        return self._body

    @property
    @ensure_design_is_active
    @min_backend_version(24, 2, 0)
    def shape(self) -> TrimmedSurface:
        """Underlying trimmed surface of the face.

        If the face is reversed, its shape is a ``ReversedTrimmedSurface`` type, which handles the
        direction of the normal vector to ensure it is always facing outward.
        """
        if self._shape is None:
            self._grpc_client.log.debug("Requesting face properties from server.")

            geometry = self._grpc_client.services.faces.get_surface(
                id=self.id, surface_type=self.surface_type
            ).get("surface")
            response = self._grpc_client.services.faces.get_box_uv(id=self.id).get("uv_box")
            u_comp, v_comp = response.get("u"), response.get("v")
            box_uv = BoxUV(Interval(*u_comp), Interval(*v_comp))

            self._shape = (
                ReversedTrimmedSurface(geometry, box_uv)
                if self.is_reversed
                else TrimmedSurface(geometry, box_uv)
            )
        return self._shape

    @property
    def surface_type(self) -> SurfaceType:
        """Surface type of the face."""
        return self._surface_type

    @property
    @ensure_design_is_active
    def area(self) -> Quantity:
        """Calculated area of the face."""
        self._grpc_client.log.debug("Requesting face area from server.")
        return self._grpc_client.services.faces.get_area(id=self.id).get("area")

    @property
    @ensure_design_is_active
    def edges(self) -> list[Edge]:
        """List of all edges of the face."""
        from ansys.geometry.core.designer.edge import CurveType

        self._grpc_client.log.debug("Requesting face edges from server.")
        response = self._grpc_client.services.faces.get_edges(id=self.id)
        return [
            Edge(
                edge.get("id"),
                CurveType(edge.get("curve_type")),
                self._body,
                self._grpc_client,
                edge.get("is_reversed"),
            )
            for edge in response.get("edges")
        ]

    @property
    @ensure_design_is_active
    def loops(self) -> list[FaceLoop]:
        """List of all loops of the face."""
        from ansys.geometry.core.designer.edge import CurveType

        self._grpc_client.log.debug("Requesting face loops from server.")
        response_loops = self._grpc_client.services.faces.get_loops(id=self.id).get("loops")
        loops = []
        for response_loop in response_loops:
            type = FaceLoopType(response_loop.get("type"))
            length = response_loop.get("length")
            min = response_loop.get("min_corner")
            max = response_loop.get("max_corner")
            edges = []
            for edge_id in response_loop.get("edges"):
                response_edge = self._grpc_client.services.edges.get_edge(id=edge_id)
                edges.append(
                    Edge(
                        response_edge.get("id"),
                        CurveType(response_edge.get("curve_type")),
                        self._body,
                        self._grpc_client,
                        response_edge.get("is_reversed"),
                    )
                )
            loops.append(
                FaceLoop(type=type, length=length, min_bbox=min, max_bbox=max, edges=edges)
            )

        return loops

    @property
    @min_backend_version(25, 2, 0)
    def color(self) -> str:
        """Get the current color of the face."""
        if self._color is None and self.body.is_alive:
            # Assigning default value first
            self._color = DEFAULT_COLOR

            # If color is not cached, retrieve from the server
            response = self._grpc_client.services.faces.get_color(id=self.id)

            # Return if valid color returned
            if response.get("color"):
                self._color = mcolors.to_hex(response.get("color"), keep_alpha=True)
            else:
                self._color = DEFAULT_COLOR

        return self._color

    @property
    def opacity(self) -> float:
        """Get the opacity of the face."""
        opacity_hex = self._color[7:]
        return int(opacity_hex, 16) / 255 if opacity_hex else 1

    @color.setter
    def color(self, color: str | tuple[float, float, float]) -> None:
        self.set_color(color)

    @opacity.setter
    def opacity(self, opacity: float) -> None:
        self.set_opacity(opacity)

    @property
    @min_backend_version(25, 2, 0)
    def bounding_box(self) -> BoundingBox:
        """Get the bounding box for the face."""
        self._grpc_client.log.debug(f"Getting bounding box for {self.id}.")
        response = self._grpc_client.services.faces.get_bounding_box(id=self.id)
        return BoundingBox(
            response.get("min_corner"), response.get("max_corner"), response.get("center")
        )

    @check_input_types
    @min_backend_version(25, 2, 0)
    def set_color(self, color: str | tuple[float, float, float]) -> None:
        """Set the color of the face."""
        self._grpc_client.log.debug(f"Setting face color of {self.id} to {color}.")
        color = convert_color_to_hex(color)
        response = self._grpc_client.services.faces.set_color(id=self.id, color=color)
        if response.get("success"):
            self._color = color
        else:  # pragma: no cover
            raise GeometryRuntimeError(f"Failed to set color {color} for face {self.id}. ")

    @check_input_types
    @min_backend_version(25, 2, 0)
    def set_opacity(self, opacity: float) -> None:
        """Set the opacity of the face."""
        self._grpc_client.log.debug(f"Setting face color of {self.id} to {opacity}.")
        opacity = convert_opacity_to_hex(opacity)

        new_color = self._color[0:7] + opacity
        self.set_color(new_color)

    @ensure_design_is_active
    def normal(self, u: float = 0.5, v: float = 0.5) -> UnitVector3D:
        """Get the normal direction to the face at certain UV coordinates.

        Parameters
        ----------
        u : float, default: 0.5
            First coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.
        v : float, default: 0.5
            Second coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.

        Returns
        -------
        UnitVector3D
            :class:`UnitVector3D` object evaluated at the given U and V coordinates.
            This :class:`UnitVector3D` object is perpendicular to the surface at the
            given UV coordinates.

        Notes
        -----
        To properly use this method, you must handle UV coordinates. Thus, you must
        know how these relate to the underlying Geometry service. It is an advanced
        method for Geometry experts only.
        """
        try:
            return self.shape.normal(u, v)
        except GeometryRuntimeError:  # pragma: no cover
            # Only for versions earlier than 24.2.0 (before the introduction of the shape property)
            self._grpc_client.log.debug(f"Requesting face normal from server with (u,v)=({u},{v}).")
            return self._grpc_client.services.faces.get_normal(id=self.id, u=u, v=v).get("normal")

    @ensure_design_is_active
    def point(self, u: float = 0.5, v: float = 0.5) -> Point3D:
        """Get a point of the face evaluated at certain UV coordinates.

        Parameters
        ----------
        u : float, default: 0.5
            First coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.
        v : float, default: 0.5
            Second coordinate of the 2D representation of a surface in UV space.
            The default is ``0.5``, which is the center of the surface.

        Returns
        -------
        Point3D
            :class:`Point3D` object evaluated at the given UV coordinates.

        Notes
        -----
        To properly use this method, you must handle UV coordinates. Thus, you must
        know how these relate to the underlying Geometry service. It is an advanced
        method for Geometry experts only.
        """
        try:
            return self.shape.evaluate_proportion(u, v).position
        except GeometryRuntimeError:  # pragma: no cover
            # Only for versions earlier than 24.2.0 (before the introduction of the shape property)
            self._grpc_client.log.debug(f"Requesting face point from server with (u,v)=({u},{v}).")
            return self._grpc_client.services.faces.evaluate(id=self.id, u=u, v=v).get("point")

    @ensure_design_is_active
    def create_isoparametric_curves(
        self, use_u_param: bool, parameter: float
    ) -> list[TrimmedCurve]:
        """Create isoparametic curves at the given proportional parameter.

        Typically, only one curve is created, but if the face has a hole, it is possible that
        more than one curve is created.

        Parameters
        ----------
        use_u_param : bool
            Whether the parameter is the ``u`` coordinate or ``v`` coordinate. If ``True``,
            it is the ``u`` coordinate. If ``False``, it is the ``v`` coordinate.
        parameter : float
            Proportional [0-1] parameter to create the one or more curves at.

        Returns
        -------
        list[TrimmedCurve]
            list of curves that were created.
        """
        self._grpc_client.log.debug(
            f"Creating isoparametric curves on face {self.id} with parameter={parameter}"
            f" and use_u_param={use_u_param}."
        )
        response = self._grpc_client.services.faces.create_iso_parametric_curve(
            id=self.id, use_u_param=use_u_param, parameter=parameter
        )

        trimmed_curves = []
        for curve in response.get("curves"):
            trimmed_curves.append(
                TrimmedCurve(
                    curve.get("geometry"),
                    curve.get("start"),
                    curve.get("end"),
                    curve.get("interval"),
                    curve.get("length"),
                    self._grpc_client,
                )
            )

        return trimmed_curves

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def setup_offset_relationship(
        self, other_face: "Face", set_baselines: bool = False, process_adjacent_faces: bool = False
    ) -> bool:
        """Create an offset relationship between two faces.

        Parameters
        ----------
        other_face : Face
            The face to setup an offset relationship with.
        set_baselines : bool, default: False
            Automatically set baseline faces.
        process_adjacent_faces : bool, default: False
            Look for relationships of the same offset on adjacent faces.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        result = self._commands_stub.FaceOffset(
            FaceOffsetRequest(
                face1=self._grpc_id,
                face2=other_face._grpc_id,
                set_baselines=set_baselines,
                process_adjacent_faces=process_adjacent_faces,
            )
        )

        return result.success

    @graphics_required
    def tessellate(self, tess_options: TessellationOptions | None = None) -> "pv.PolyData":
        """Tessellate the face and return the geometry as triangles.

        Parameters
        ----------
        tess_options : TessellationOptions | None, default: None
            A set of options to determine the tessellation quality.

        Returns
        -------
        ~pyvista.PolyData
            :class:`pyvista.PolyData` object holding the face.

        Notes
        -----
        The tessellation options are ONLY used if the face has not been tessellated before.
        If the face has been tessellated before, the stored tessellation is returned.
        """
        # If tessellation has not been called before... call it
        if self._body._template._tessellation is None:
            self._body.tessellate(tess_options=tess_options)

        # Search the tessellation of the face - if it exists
        # ---> We need to used the last element of the ID since we are looking inside
        # ---> the master body tessellation.
        red_id = self.id.split("/")[-1]
        mb_pdata = self.body._template._tessellation.get(red_id)
        if mb_pdata is None:  # pragma: no cover
            raise ValueError(f"Face {self.id} not found in the tessellation.")

        # Return the stored PolyData
        return mb_pdata.transform(self.body.parent_component.get_world_transform(), inplace=False)

    @graphics_required
    def plot(
        self,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        **plotting_options: dict | None,
    ) -> None:
        """Plot the face.

        Parameters
        ----------
        screenshot : str, default: None
            Path for saving a screenshot of the image that is being represented.
        use_trame : bool, default: None
            Whether to enable the use of `trame <https://kitware.github.io/trame/index.html>`_.
            The default is ``None``, in which case the
            ``ansys.tools.visualization_interface.USE_TRAME`` global setting is used.
        use_service_colors : bool, default: None
            Whether to use the colors assigned to the face in the service. The default
            is ``None``, in which case the ``ansys.geometry.core.USE_SERVICE_COLORS``
            global setting is used.
        **plotting_options : dict, default: None
            Keyword arguments for plotting. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.

        """
        # lazy import here to improve initial module loading time
        import ansys.geometry.core as pyansys_geometry
        from ansys.geometry.core.plotting import GeometryPlotter
        from ansys.tools.visualization_interface.types.mesh_object_plot import (
            MeshObjectPlot,
        )

        use_service_colors = (
            use_service_colors
            if use_service_colors is not None
            else pyansys_geometry.USE_SERVICE_COLORS
        )

        mesh_object = self if use_service_colors else MeshObjectPlot(self, self.tessellate())
        pl = GeometryPlotter(use_trame=use_trame, use_service_colors=use_service_colors)
        pl.plot(mesh_object, **plotting_options)
        pl.show(screenshot=screenshot, **plotting_options)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing server-version agnostic conversions."""

from pint import Quantity

from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Distance, Measurement


def from_measurement_to_server_length(input: Measurement) -> float:
    """Convert a measurement to a length value.

    Parameters
    ----------
    value : Measurement
        Measurement value.

    Returns
    -------
    float
        Length value in server-defined units. By default, meters.
    """
    return input.value.m_as(DEFAULT_UNITS.SERVER_LENGTH)


def from_measurement_to_server_angle(input: Measurement) -> float:
    """Convert a measurement to an angle value.

    Parameters
    ----------
    value : Measurement
        Measurement value.

    Returns
    -------
    float
        Angle value in server-defined units. By default, radians.
    """
    return input.value.m_as(DEFAULT_UNITS.SERVER_ANGLE)


def to_distance(value: float | int) -> Distance:
    """Convert a server value to a Distance object.

    Parameters
    ----------
    value : float | int
        Value to convert.

    Returns
    -------
    Distance
        Converted distance.

    Notes
    -----
    The value is converted to a Distance object using the default server length unit.
    The value should represent a length in the server's unit system.
    """
    return Distance(value, DEFAULT_UNITS.SERVER_LENGTH)


def to_area(value: float | int) -> Quantity:
    """Convert a server value to an area object.

    Parameters
    ----------
    value : float | int
        Value to convert.

    Returns
    -------
    Quantity
        Converted area.

    Notes
    -----
    The value is converted to a Quantity object using the default server area unit.
    The value should represent an area in the server's unit system.
    """
    return Quantity(value, DEFAULT_UNITS.SERVER_AREA)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the data structure for material and material properties."""

from collections.abc import Sequence

from beartype import beartype as check_input_types
from pint import Quantity

from ansys.geometry.core.materials.property import MaterialProperty, MaterialPropertyType


class Material:
    """Provides the data structure for a material.

    Parameters
    ----------
    name: str
        Material name.
    density: ~pint.Quantity
        Material density.
    additional_properties: Sequence[MaterialProperty], default: None
        Additional material properties.
    """

    @check_input_types
    def __init__(
        self,
        name: str,
        density: Quantity,
        additional_properties: Sequence[MaterialProperty] | None = None,
    ):
        """Initialize the ``Material`` class."""
        self._name = name
        self._density = MaterialProperty(MaterialPropertyType.DENSITY, "Density", density)
        if not additional_properties:
            additional_properties = []

        # Add the density to the properties list
        additional_properties.append(self._density)
        self._properties = {}
        for property in additional_properties:
            self._properties[property.type] = property

    @property
    def properties(self) -> dict[MaterialPropertyType, MaterialProperty]:
        """Dictionary of the material property type and material properties."""
        return self._properties

    @property
    def name(self) -> str:
        """Material name."""
        return self._name

    @check_input_types
    def add_property(self, type: MaterialPropertyType, name: str, quantity: Quantity) -> None:
        """Add a material property to the ``Material`` class.

        Parameters
        ----------
        type : MaterialPropertyType
            Material property type.
        name: str
            Material name.
        quantity: ~pint.Quantity
            Material value and unit.
        """
        property = MaterialProperty(type, name, quantity)
        self._properties[property.type] = property
